//                              -*- Mode: C++ -*-
//: EAHypTraAnalyzer.C
//
//
// Author: Mauro Villa   
// ECAL group 
// HERA-B experiment, Desy
// e-mail: villa@bo.infn.it
//
// Created 2/5/97
// Last Update ........
//


#include "care/EMClusterTree.hh"
#include "care/EAHypTraAnalyzer.hh"
#include "arte/ArteMessage.hh"
#include "care/ECStrategyFactory.hh"
#include "care/ESCorCOGPos.hh"
#include "arte/ArteProto.hh"
#include "care/EMHitCell.hh"
#include "care/EMTrackOnEcal.hh"
#include "care/CmdCareLevel.hh"

#undef WRITE


//////////////////// EAHypTraAnalyzer Class Definitions ////////////////

EAHypTraAnalyzer::EAHypTraAnalyzer(vector<const EMCluster*> & hypShow,
				   vector<const EMCluster*> & hypSign) : 
  EAHypAnalyzer(hypShow), hypSignals(&hypSign) {

    nextRCCL = 1;  // point to the first entry in the RCCL bank
    nextClus = 0; 
    if(hypShowers->begin()!=hypShowers->end())
      nextClus = hypShowers->begin(); // point to the first cluster in hypShowers
    passed = false;  // set the flag that the cluster is not passed yet
}

EAHypTraAnalyzer::~EAHypTraAnalyzer(){}

//--
// ClusterMember Analyzer
void EAHypTraAnalyzer::doAnalysis(const EMClusterMember *){
  static int EAHypTraAnalyzerWarn = 0x0ECA0004;
  arteMsg<<" EAHypTraAnalyzer: Do Analysis called for a ClusterMember ";
  arteMsg.warn(EAHypTraAnalyzerWarn);
}



//--
//  analyze the Top list
void EAHypTraAnalyzer::analyzeTopList(const EMClusterTree& topList){
  doAnalysis(&topList);
}

//--
//  analyze the Top list
void EAHypTraAnalyzer::doAnalysis(const EMClusterTree* tree){
# ifdef WRITE
  tree->dump();
  cout<<"A";
# endif
  if(hypSignals->begin()!=hypSignals->end()){
    // erase all data
    hypSignals->erase(hypSignals->begin(),hypSignals->end());
  }
# ifdef WRITE
  cout<<"E";
# endif
  nextRCCL = 1; // just to be sure
  nextClus = 0; 
  if(hypShowers->begin()!=hypShowers->end()){
#   ifdef WRITE
    cout<<"I";
#   endif
    // move all showers to signals
    //copy(hypShowers->begin(),hypShowers->end(), hypSignals->begin());
    for(nextClus = hypShowers->begin(); nextClus!=hypShowers->end(); 
    	nextClus++){
      hypSignals->push_back(*nextClus);
    }
    nextClus = hypShowers->begin();
  }
  passed = false;
# ifdef WRITE
  cout<<"O"<<endl;
# endif
  // main loop on Top List
  for(const EMClusterMember* child = tree->begin(); child!=tree->end(); 
      child = child->next()){
    if( !child->isALeaf()){
      child->analyzeMe(*this); 
    }
  }
# ifdef WRITE
  cout<<"U"<<endl;
# endif
}

//--
// HitCell Analyzer
void EAHypTraAnalyzer::doAnalysis(const EMHitCell *){
  // called for hits of clusters with more than 1 hypotheses
}

 
//--
// Cluster Analyzer
void EAHypTraAnalyzer::doAnalysis(const EMCluster * cluster){
  // start testing if the current cluster is that one pointed by nextClus
  if( nextClus!=0 && cluster== (*nextClus) ){
#   ifdef WRITE
    cout<<"F"<<nextRCCL;
#   endif
    // It is the same cluster
    passed = true;
    if(cluster->getHypotheses()==1){
#   ifdef WRITE
      cout<<"U";
#   endif
      // one hypotheses only!!
      if(cluster->getTracks()==0){
	// One neutral cluster
        // All has been done before if level=2
	if( CmdCareLevel::level()>3){
	  updateOldEntry(cluster, nextRCCL);
	}
        rcLinks.fillNewEntry(cluster, nextRCCL);
      } else {
	// one track... something must be done to store track links
	updateOldEntry(cluster, nextRCCL);
        rcLinks.fillNewEntry(cluster, nextRCCL);
      }
      hypSignals->operator[](nextRCCL.index()-1) = cluster;
    } else {
#     ifdef WRITE
      cout<<"M";
#     endif
      // more than one hypothesis
      // the first single hypothesis must be updated and the following
      // entered in the RCCL bank (if they exist)
      const EMClusterMember* child = cluster->begin();
      if( child->isALeaf()){
	// I am dealing with a base cluster with more than 1 hypothesis
	const EMCluster* trueClus = cluster;
	while( trueClus->getHypotheses() == cluster->getHypotheses() &&
	       trueClus->depthLevel() > FIRSTLEVEL  &&  
	       (trueClus!= (*nextClus)))
	  trueClus = trueClus->getParent();
#       ifdef WRITE
	cout<<"uwrong!!!";
#       endif
	rcLinks.fillNewEntry(cluster, nextRCCL); /// ?? something wrong
        // updateOldEntry(trueClus, nextRCCL);
	hypSignals->operator[](nextRCCL.index()-1) = cluster;

      } else {
#       ifdef WRITE
	cout<<"l";
#       endif
	// more than one or 0 cluster hypotheses: loop on COMPLEX cluster
	for(; child!=0;  child = child->next() ){
	  child->analyzeMe(*this);
	}
      }
    }
    // get ready for the next cluster
    if( nextClus!=0 ){
      nextClus++;
      if( nextClus == hypShowers->end()){
	nextClus = 0;
      }
    }
    nextRCCL = nextRCCL.index()+1;
    passed = false;
#   ifdef WRITE
    cout<<"New"<<nextRCCL;
#   endif
  } else if( cluster->getHypotheses()!=1 ){
#   ifdef WRITE
    cout<<"L";
#   endif
    // more than one or 0 cluster hypotheses: loop on cluster
    /// warn: here I should test for many tracks in one base cluster...
    const EMClusterMember* child = cluster->begin();
    if( child->isALeaf() && cluster->getHypotheses()>1){
#   ifdef WRITE
      cout<<" WRONG!!!!!!!!! ";
#   endif
      // I am dealing with a base cluster with more than 1 hypothesis
      const EMCluster* trueClus = cluster;
      while( trueClus->getHypotheses() == cluster->getHypotheses() &&
             trueClus->depthLevel()>FIRSTLEVEL && trueClus != *nextClus)
	     trueClus = trueClus->getParent();
      if( passed ){
#       ifdef WRITE
        cout<<"R"<<nextRCCL;
#       endif
        // update the current bank
        updateOldEntry(trueClus, nextRCCL);
	rcLinks.fillNewEntry(cluster, nextRCCL);
	*nextClus = trueClus; 
	hypSignals->operator[](nextRCCL.index()-1) = trueClus;
        passed = false;
      } else {
#       ifdef WRITE
        cout<<"B";
#       endif
        // enter another RCCL entry
        fillNewEntry(cluster);
        rcLinks.fillNewEntry(cluster, ArtePointer<RCCL>::nrow() );
	hypSignals->operator[](ArtePointer<RCCL>::nrow()-1) = cluster;
      }
      //      hypSignals->push_back(cluster);
    } else {
      // more than one or 0 cluster hypotheses: loop on COMPLEX cluster
      for(; child!=0;  child = child->next() ){
        child->analyzeMe(*this);
      }
    }	
  } else {
#   ifdef WRITE
    cout<<"O"<<nextRCCL;
#   endif
    // here with exactly one cluster hypothesis
    if( passed ){
#     ifdef WRITE
      cout<<"R";
#     endif
      // update the current bank
      updateOldEntry(cluster, nextRCCL);
      rcLinks.fillNewEntry(cluster, nextRCCL);
      hypSignals->operator[](nextRCCL.index()-1) = cluster;
      passed = false;
    } else {
#     ifdef WRITE
      cout<<"Bx";
#     endif
      // enter another RCCL entry
      fillNewEntry(cluster);
      rcLinks.fillNewEntry(cluster, ArtePointer<RCCL>::nrow() );
      // hypSignals->operator[](ArtePointer<RCCL>::nrow()-1) = cluster;
      hypSignals->push_back(cluster);
    }
  }
}

//--
// HitCell Analyzer
void EAHypTraAnalyzer::doAnalysis(const EMTrackOnEcal *){}
