//                              -*- Mode: C++ -*-
//: EMCluster.C
//
//
// Author: Mauro Villa   
// ECAL group 
// HERA-B experiment, Desy
// e-mail: villa@bo.infn.it
//
// Created 3/4/97
// Last Update ........
//

#include "care/EMCluster.hh"
#include "care/EAClusterAnalyzer.hh"
#include "care/EDClusterModifier.hh"
#include "arte/ArteMessage.hh"

//////////////  Class EMCluster definitions //////////////////////////////

//-- Constructors
//   The construction of EMCluster is done referring to some
//   parent. The depth level is adjusted accordingly.
EMCluster::EMCluster(EMCluster* parent, EMClusterMember* child) :
  EMClusterMember(parent), hyp() {
    firstChild = child;
    lastChild = child;
    child->parentCluster = this;
}


// Default constructor
EMCluster::EMCluster() : EMClusterMember(), firstChild(0), lastChild(0), 
    hyp() {}


//-- Destructor
//   A virtual destructor is needed
EMCluster::~EMCluster(){
  // start destroying each object in the list
  // then remove the list
  EMClusterMember * cp = firstChild, *childp;
  while( cp!=0){
    childp = cp->nextClusterMember;
    cp->~ClusterMember();
    cp = childp;
  }
}


//--
// Add a Child to the list
EMClusterMember * EMCluster::add(EMClusterMember* child){
  // start adding this child to the list;
  // assume the child do not belong to any cluster
  child->nextClusterMember = 0;
  child->previousClusterMember = lastChild;
  lastChild->nextClusterMember = child;
  lastChild = child;
  // set the new parent
  child->parentCluster = this;
  // set the depth level of the child and its children
  child->adjust_depths();
  return this;
}


//--
// Move a Child from a list to this list
void EMCluster::move(EMClusterMember * pChild){
  Energy_t childEn = pChild->maxEnergy();  // MaxEnergy of the child to Move
  // Find position to insert child
  EMClusterMember* where = firstChild;
  //
  while(where!=0){
    if( where->maxEnergy() < childEn) break; // a position has been found
    where = where->nextClusterMember;
  }
  // Now I have to insert child "pChild" of SourceCluster into this cluster
  // in position where
  // start removing from SourceCluster
  if( pChild->previousClusterMember!=0 ){
    //
    pChild->previousClusterMember->nextClusterMember = 
      pChild->nextClusterMember;
  }
  if( pChild->nextClusterMember!=0 ){
    //
    pChild->nextClusterMember->previousClusterMember = 
      pChild->previousClusterMember;
  }
  if(pChild->parentCluster!=0){
    //
    EMCluster*  pC = pChild->parentCluster;
    if( pC->firstChild == pChild )
      pC->firstChild = pChild->nextClusterMember;
    if( pC->lastChild == pChild )
      pC->lastChild = pChild->previousClusterMember;
  }
  if(where==0){ //move to the end 
    //
    lastChild->nextClusterMember = pChild;
    pChild->previousClusterMember = lastChild;
    pChild->nextClusterMember = 0;
    lastChild = pChild;
  } else {
    //
    pChild->nextClusterMember = where;
    pChild->previousClusterMember = where->previousClusterMember; 
    where->previousClusterMember = pChild;
    if(where==firstChild)
      firstChild = pChild;
  }
  //
  // Update child parameters 
  pChild->parentCluster = this;
  // set the depth level of the child and its children
  pChild->adjust_depths();  
}
 

//-- 
//  Start an uninitialized cluster with the Child passed
void EMCluster::start(EMClusterMember* child){
  // 
  // start updating list pointers of the current object
  firstChild = child;
  lastChild = child;
  // update info current node
  depth = child->depthLevel();
  // 
  parentCluster = child->parentCluster;
  child->parentCluster = this;
  // update links to and from current node
  // links at other nodes
  // 
  nextClusterMember = child->nextClusterMember;
  previousClusterMember = child->previousClusterMember;
  // links to this node
  if( nextClusterMember!=0 )
    nextClusterMember->previousClusterMember = this;
  // 
  if( previousClusterMember!=0 )
    previousClusterMember->nextClusterMember = this;
  // parent links
  if( parentCluster!=0 ){
    if( parentCluster->firstChild==child)
      parentCluster->firstChild = this;
    // 
    if( parentCluster->lastChild==child)
      parentCluster->lastChild = this;
    // 
  }
  // put to zero links of the child
  child->nextClusterMember = 0;
  child->previousClusterMember = 0;
  // now all links are ok
  // update depth level of child
  // 
  child->adjust_depths();
  // Set the number of Hypotheses to zero
  hyp.reset();
}




// Status access functions
//--
// Return a flag to tell that this EMClusterMember is a leaf 
bool EMCluster::isALeaf() const {
  return false;
}

//--
// Return the maximum energy measured in the cluster's (and sub clusters) 
// cells or the Energy of the Cell itself. 
Energy_t  EMCluster::maxEnergy() const {
  // Since the list is ordered in energy by the max cell energy,
  // the Maximum Energy is given by that of the first child in the list
  // if it exists;
  Energy_t maxEn = 0.0;
  EMClusterMember* pChild = firstChild;
  if(pChild!=0 )
    maxEn = pChild->maxEnergy();
  return maxEn;
}

//--
// After operation on cluster such as reordering, merging or other
// It is possible that the depths are not correct. With this function
// the depth of the EMClusterMember and all its children are adjusted. 
void EMCluster::adjust_depths(){
  // start adjust its depth
  EMClusterMember::adjust_depths();
  // continue adjusting children depths
  for(EMClusterMember* child = firstChild; 
      child!=0;  child = child->nextClusterMember)
      child->adjust_depths();
}

//--
// Analysis and traversing of the cluster tree is done by EAClusterAnalyzers
// that are object with definite analyzing properties. 
// EMClusterMembers MUST ignore these properties: in this way a loose coupling
// between EMClusterMembers and EAClusterAnalyzers is achieved and both
// can vary independently during program evolution.
// Examples of EAClusterAnalyzers are those that should return
// total energy of a cluster, the Center of Gravity, the Number of Cells.
void EMCluster::analyzeMe(EAClusterAnalyzer & analyzer) const {
  //
  analyzer.doAnalysis(this);
}

void EMCluster::modifyMe(EDClusterModifier & modifier){
  //
  modifier.doModify(this);
}


//--
//  Dump for debugging
void EMCluster::dump() const {
  for(int i=0; i<depthLevel(); i++)
    arteMsg<<"|";
  arteMsg<<"->Cluster ";
  hyp.dump();
  for(EMClusterMember* child = firstChild; 
      child != 0; 
      child = child->nextClusterMember){
    child->dump();
    if((const EMCluster *)child->getParent()!=this){
      for(int i=0; i<depthLevel(); i++)
	arteMsg<<"| ";
      arteMsg<<"^ Bad Parent EMCluster";
      arteMsg.allp();
    }
  }
}


//-- Reset all info in the cluster
void  EMCluster::reset(){
  depth = TOPLEVEL;
  parentCluster= 0;
  nextClusterMember = 0;
  previousClusterMember = 0;
  firstChild = 0;
  lastChild = 0;
  // Set the number of Hypotheses to zero
  hyp.reset();
}


//--
// Move a Child from its' parent list to this cluster and put it at the
// end
EMClusterMember* EMCluster::moveAtEnd(EMClusterMember* pChild){
  Energy_t childEn = pChild->maxEnergy();  // MaxEnergy of the child to Move
  // Find position to insert child
  EMClusterMember* where = firstChild;
  while(where!=0){
    if( where->maxEnergy() < childEn) break; // a position has been found
    where = where->nextClusterMember;
  }
  // Now I have to insert child "pChild" into this cluster
  // at the and

  // start removing from its parentCluster
  if( pChild->previousClusterMember!=0 ){
    pChild->previousClusterMember->nextClusterMember = 
      pChild->nextClusterMember;
  }
  if( pChild->nextClusterMember!=0 ){
     pChild->nextClusterMember->previousClusterMember = 
       pChild->previousClusterMember;
  }
  // update links of the parent cluster
  if(pChild->parentCluster!=0){
    EMCluster*  pC = pChild->parentCluster;
    if( pC->firstChild == pChild )
      pC->firstChild = pChild->nextClusterMember;
    if( pC->lastChild == pChild )
      pC->lastChild = pChild->previousClusterMember;
  }
  // now the child is free from its parent cluster.
  // Add it at the end
  if(lastChild!=0){
    lastChild->nextClusterMember = pChild;
  }
  pChild->previousClusterMember = lastChild;
  pChild->nextClusterMember = 0;
  lastChild = pChild;
  // Update child parameters 
  pChild->parentCluster = this;
  // set the depth level of the child and its children
  pChild->adjust_depths();  

  // return the child
  return pChild;
}  
 
