//                              -*- Mode: C++ -*-
//: ESPeakPos.C
//
//
// Author: Mauro Villa   
// ECAL group 
// HERA-B experiment, Desy
// e-mail: villa@bo.infn.it
//
// Created 12/6/97
// Last Update ........
//

#include "arte/arte.hh"
#include "care/caredata.hh"
#include "care/ESPeakPos.hh"
#include "care/EMClusterTree.hh"
#include "care/ECCellID.hh"


/*
CLASS
  ESPeakPos
  */

////////////////////// static data definitions  ////////////////////////
// here there are stored the variance on x and y due to the cell size
// the number you can see are evaluated in this way:
// var[i] = (11.15/i)^2/12  for i=1,2,5
Coord_t ESPeakPos::varsx[] = {0.0, 10.3602, 2.59005, 0.0, 0.0, 0.414408};
Coord_t ESPeakPos::varsy[] = {0.0, 10.3602, 2.59005, 0.0, 0.0, 0.414408};

////////////////////// static member function definition //////////////
// initialize with true cell size
void ESPeakPos::initialize(){
  float widx = ECalInfo.xMsize;
  float widy = ECalInfo.yMsize;
  float varx = widx*widx/12;
  float vary = widy*widy/12;
  varsx[0] = 0.0;
  varsy[0] = 0.0;
  for(int i=1; i<6; i++){
    varsx[i] = varx/(i*i);
    varsy[i] = vary/(i*i);
  }
}


////////////////////// member function definitions ///////////////////

ESPeakPos::ESPeakPos() : ESPosStrategy() , hitt(0) {};
ESPeakPos::~ESPeakPos(){};


//-- Behavioural operations
//   evaluate the position for a cluster
void ESPeakPos::evaluatePos(const EMCluster* clus) {
  // This function analyzes the pointed cluster to find the most
  // energetic cell in the cluster and evaluate its position and width
  // It fills in the local part data that will be used for a rough estimation
  // of the cluster position

  hitt = 0; // set the Hit Cell pointer to null
  
  clus->analyzeMe(*this); // Look for the most energetic cell in the cluster

  if(hitt!=0){      // ok the peak has been found
    //
    unsigned int ncl = hitt->ncl();
    //
    posData.setPoint(hitt->x(), hitt->y(), ECalInfo.ecalp->z1);
    posData.setVar( varsx[ncl], 0.0, varsy[ncl]);
    posData.setEnergy( hitt->energy()); // energy
    posData.setSqEnErr( hitt->sqerror()); // Error on energy
    posData.setWidth(1);        // one cell considered
    posData.setType( Peak );      // peak algorithm applied
    posData.setCell( hitt );             // peak position
  } else {
    posData.reset();        // reset the data on position
  }
} 


//   search for the most energetic cell in a cluster
void ESPeakPos::doAnalysis(const EMCluster* clus) {
  // the peak is always present in the first ClusterMember of this cluster
  // however if the peak is not found as the first take the first hit cell.
  EMClusterMember* child = clus->begin();
  while( child!=0 && hitt==0){
    child->analyzeMe(*this);
    child = child->next();
  }
} 

//   Analysis of a cell
void ESPeakPos::doAnalysis(const EMHitCell* hitT) {
  // when here this cell is the peak we are looking for
  hitt = hitT;
} 

//   Analysis of a tree
void ESPeakPos::doAnalysis(const EMClusterTree* tree) {
  const EMCluster* cluster = tree;
  doAnalysis(cluster);
} 

//   Nothing to do for ClusterMember
void ESPeakPos::doAnalysis(const EMClusterMember*) {
} 


//--
// TrackOnEcal Analyzer
void ESPeakPos::doAnalysis(const EMTrackOnEcal*){
}


