#include <algorithm>
#include <stdio.h>
#include "poirot/Cluster.hh"
#include "poirot/TrackDescription.hh"
#include "poirot/ReconstructedHit.hh"

#ifdef __OTR_USE_ROOT__
#include <TH1.h>
#include <TProfile.h>
#endif

//__________________________________________________________________
// poirotCluster
//
// Container for:
// * 3 OTrSectorclusters (but not owner)
// Functionality:
// * Reconstruction of one spacepoint
//
// Author: Wouter Hulsbergen 
//__________________________________________________________________

float poirotCluster::fgMinChi2Probability = 0.001 ;

poirotCluster*
poirotCluster::Create(poirotSectorCluster* plusIt, 
		      poirotSectorCluster* minusIt) 
{
  float sindif 
    //sin(plusIt->GetStereoAngle()-minusIt->GetStereoAngle()) ;
    = plusIt->GetSinStereoAngle()*minusIt->GetCosStereoAngle() -
    plusIt->GetCosStereoAngle()*minusIt->GetSinStereoAngle() ;
  float tx = (plusIt->GetSinStereoAngle()*minusIt->GetRawTx() - 
	      minusIt->GetSinStereoAngle()*plusIt->GetRawTx())/sindif ;
  float ty = (plusIt->GetCosStereoAngle()*minusIt->GetRawTx() - 
	      minusIt->GetCosStereoAngle()*plusIt->GetRawTx())/sindif ;
  
  poirotCluster* sc = 0 ;
  if(plusIt->WithinYAcceptance(tx,ty) &&
     minusIt->WithinYAcceptance(tx,ty)) {
    sc = new poirotCluster() ;
    sc->SetTx(tx) ;
    sc->SetTy(ty) ;
    sc->SetPlusCluster(plusIt) ;
    sc->SetMinusCluster(minusIt) ;
  }
  return sc ;
}

poirotCluster::poirotCluster()
  : poirotTrackDescription2D(), fXDif(99), fSectorClusters(3,NULL) {}
  
poirotCluster::poirotCluster(const poirotCluster& rhs) 
  : poirotReconstructedTrackInterface(rhs),
    poirotTrackDescription2D(rhs)
{
  fXDif = rhs.fXDif ;
  fSectorClusters = rhs.fSectorClusters ;
}

poirotCluster::~poirotCluster() {}

bool
poirotCluster::SetZeroIfValid(poirotSectorCluster* zeroClus) {
  bool good(false) ;
  if(zeroClus->WithinYAcceptance(GetTx(),GetTy())) {
    float xdif = zeroClus->GetXDif(GetTx()) ;
    if((!GetZeroCluster() && 
	fabs(xdif) < 
	(GetPlusCluster()->GetWirePitch()/
	 GetPlusCluster()->GetCosStereoAngle() +
	 zeroClus->GetWirePitch())/2) ||
       (GetZeroCluster() && fabs(xdif) < fabs(fXDif))) {
      fXDif = xdif ;
      SetZeroCluster(zeroClus) ;
      good = true ;
    }
  }
#ifdef __OTR_USE_ROOT__
  static TH1F* xdifH1 = new TH1F("xdifB","",500,-3,3) ;
  if(zeroClus->WithinYAcceptance(GetTx(),GetTy())) {
    float x = zeroClus->GetXDif(GetTx())/
      zeroClus->GetWirePitch() ;
    xdifH1->Fill(x) ;
  }
#endif //__OTR_USE_ROOT__
  return good ;
}

bool
poirotCluster::Merge(poirotCluster& rhs)
{

  // The idea is this: If clusters overlap in more than one
  // stereolayer, only one of them can be real or they form one
  // cluster. This is of course not completely true, but in any case
  // almost completely.

  // Merge clusters that overlap in zero && plus or in zero && minus

  // 04/08/99 I give up those terrible procedures that try to merge
  // clusters. The new implementation will be: If clusters have only
  // one sector-cluster in common, we delete neither of them. If they
  // have two in common, we delete the worst one.

  int ncommon(0) ;
  if(GetZeroCluster() ==rhs.GetZeroCluster()) ncommon++;
  if(GetPlusCluster() ==rhs.GetPlusCluster()) ncommon++;
  if(GetMinusCluster()==rhs.GetMinusCluster()) ncommon++;

  bool merged = (ncommon>1) ;
  return merged ;

}

bool
poirotCluster::MixedPitch() const {
  bool mixed = false ;
  if(GetHits().size()>0) {
    HitContainer::const_iterator hitIt = GetHits().begin() ;
    poirotGede::EPitch gedepitch =
      (*hitIt)->GetMeasuredHit()->GetGede()->GetEPitch() ;
    for(; hitIt != GetHits().end(); hitIt++)
      if(gedepitch != (*hitIt)->GetMeasuredHit()->GetGede()->GetEPitch()) {
	mixed = true ;
	break ;
      }
  }
  return mixed ;
}

enum FitMode { kNormal=0, kCompareChi2=1, kCheckChi2=3 };

extern "C" float prob_(const float&, const int&) ;

bool
poirotCluster::OneFitIteration(int fitmode)
{
  HepDouble chi2 = 0;
  static HepVector dChi2dQ(GetDim(),0) ;
  static HepSymMatrix d2Chi2dQ2(GetDim(),0) ;
  for(int row=0; row<GetDim(); row++) {
    dChi2dQ[row]=0;
    for(int col=0; col<=row; col++)
      d2Chi2dQ2[row][col]=0;
  }
  poirotNodeProjection* nodeProjection;
  for(HitContainer::iterator hitIt=GetHits().begin() ; 
      hitIt!=GetHits().end(); hitIt++) {
    nodeProjection = (*hitIt)->GetNodeProjection() ;
    nodeProjection->Project(this) ;
    chi2      += nodeProjection->GetChi2() ;
    dChi2dQ   += nodeProjection->GetdChi2dQ() ;
    d2Chi2dQ2 += nodeProjection->Getd2Chi2dQ2() ;
  }
  int ierr;
  HepSymMatrix d2Chi2dQ2Inv = d2Chi2dQ2.inverse(ierr) ;
  if(ierr) {
    printf("Matrix inversion error in poirotCluster::OneFitIteration()") ;
    return false ;
  }
  chi2 += -d2Chi2dQ2Inv.similarity(dChi2dQ)/2 ;// This is the expected change in chi2
  switch (fitmode)
    {
    case kCompareChi2: 
      if(chi2>GetChi2()) return false ;
      break;
    case kCheckChi2:
      if(!ValidChi2(chi2)) return false ;
      break;
    }
  SetQ(GetQ() - d2Chi2dQ2Inv*dChi2dQ ) ;
  SetCov(2*d2Chi2dQ2Inv) ;
  SetChi2(chi2) ;
  return ValidChi2() ;
}

extern "C" float prob_(const float&, const int&) ;

inline bool 
compareHitDistance(poirotReconstructedHit* lhs, poirotReconstructedHit* rhs) {
  float lhsr = lhs->GetNodeProjection()->GetRm()/lhs->GetMeasuredHit()->
    GetGede()->GetRMax() ;
  float rhsr = rhs->GetNodeProjection()->GetRm()/rhs->GetMeasuredHit()->
    GetGede()->GetRMax() ;
  return lhsr > rhsr ;
}

inline bool 
compareHitChi2(poirotReconstructedHit* lhs, poirotReconstructedHit* rhs) 
{
  return lhs->GetNodeProjection()->GetChi2()>
    rhs->GetNodeProjection()->GetChi2() ;
}

bool
poirotCluster::CreateReconstructedHits()
{
  //cout << "poirotCluster::CreateReconstructedHits()" << endl ;
  bool success(false) ;
  EraseHits();
  for(ClusterContainer::const_iterator clusIt=fSectorClusters.begin() ;
      clusIt!=fSectorClusters.end() ; clusIt++)
    for(poirotSectorCluster::HitIterator hitIt=(*clusIt)->begin(); 
	hitIt!=(*clusIt)->end() ; hitIt++) {
      poirotReconstructedHit::LRSign_t lrsign = (*clusIt)->GetLRSign(*hitIt) ;
      GetHits().reserve(GetHits().size()+1) ;// without this one, a memory leak
      GetHits().push_back(new poirotReconstructedHit(*hitIt,lrsign)) ;
    }
    
  // Sort hits in measured distance and determine lrsign one-by-one.
  sort(GetHits().begin(),GetHits().end(),compareHitDistance) ;

  for(HitContainer::iterator hitIt=GetHits().begin() ;
      hitIt!=GetHits().end() ; hitIt++) 
    if(!(*hitIt)->GetLRSignIsFixed()) {
      (*hitIt)->SetLRSign(-1.0) ;
      OneFitIteration(kNormal) ; 
      (*hitIt)->FlipLRSign() ;
      if(!OneFitIteration(kCompareChi2)) (*hitIt)->FlipLRSign() ; // Fit worse, flip back
    }
  OneFitIteration(kNormal) ; 

  // This does not really make it better ...
  for(HitContainer::iterator hitIt=GetHits().begin() ;
      hitIt!=GetHits().end(); hitIt++) (*hitIt)->UpdateInternal() ;
  OneFitIteration(kNormal) ;

  // Flip all hits once to see if we can improve and get rid of really bad hits
  success = OneFitIteration(kNormal) ; // We need this second to set the residuals right !!!
  sort(GetHits().begin(), GetHits().end(), compareHitChi2) ;
  HitContainer badHits ;
  for(HitContainer::iterator hitIt=GetHits().begin() ; 
      hitIt!=GetHits().end(); hitIt++) {
    //cout << (*hitIt)->GetChi2() << " " ;
    (*hitIt)->FlipLRSign() ;
    success = OneFitIteration(kCompareChi2) ;
    bool changed=success;
    if(!ValidChi2() && ValidNHits(*hitIt)) {
      //cout << " removed. " ;
      badHits.push_back(*hitIt) ;
      GetHits().erase(hitIt--) ;
      OneFitIteration(kNormal) ;
      changed=true ;
    } else {
      if(!success) (*hitIt)->FlipLRSign() ;
      //else cout << " lrsign changed " ;
    }
    //cout << "|||" ;
    // If you need the residuals longer, you have to take care they are up-to-date:
    if(changed) {
      OneFitIteration(kNormal) ;
      sort(hitIt+1,GetHits().end(),compareHitChi2) ;
    }
  }
  success = ValidChi2() ;
  // Give all bad hits a last chance and delete them if they fail
  for(HitContainer::iterator hitIt=badHits.begin() ;
      hitIt!=badHits.end(); hitIt++) {
    (*hitIt)->GetNodeProjection()->Project(this,true,true) ; // to set lrsign right
    GetHits().push_back(*hitIt) ;
    if(!OneFitIteration(kCheckChi2)) {
      GetHits().pop_back() ;
      delete *hitIt ;
    }
  }

  OneFitIteration(kNormal) ;

#ifdef __OTR_USE_ROOT__
  static TH1F* chi2H1 = new TH1F("scchi2H1","",1000,0,100) ;
  chi2H1->Fill(GetChi2OverNDof()) ;
  static TH1F* chi2probH1 = new TH1F("scchi2probH1","",1000,0,1) ;
  //if(fChi2 < 100*chisin_(1-fgMinChi2Prob,GetNDof()))
  chi2probH1->Fill(GetChi2Prob()) ;
  static TH1F* clusterSigmaXH1 = new TH1F("clusterSigmaXH1","",1000,0,0.1) ;
  clusterSigmaXH1->Fill(sqrt(GetCov()[0][0])*GetCentralZ()) ;
  static TH1F* clusterSigmaYH1 = new TH1F("clusterSigmaYH1","",1000,0,10.0) ;
  clusterSigmaYH1->Fill(sqrt(GetCov()[1][1])*GetCentralZ()) ;  
  if(GetChi2Prob() > 0.001) {
    static TH1F* lrsignNH1 = new TH1F("lrsignNH1","",3,-1.5,1.5) ;
    static TProfile* lrsignNPR = new TProfile("lrsignNPR","",20,-1,1) ;
    for(HitContainer::iterator hitIt=GetHits().begin() ;
	hitIt!=GetHits().end(); hitIt++) {
      // 	static TProfile* lrSignPR = new TProfile("lrSignPR","",20,-1,1) ;
      ArtePointer<HITB> hitb = (*hitIt)->GetMeasuredHit()->GetHITB() ;
      int realSign = int(hitb->q / fabs(hitb->q)) ;
      int recSign = (*hitIt)->GetLRSign() ;
      lrsignNH1->Fill(recSign*realSign) ;
      lrsignNPR->Fill(hitb->q/((*hitIt)->GetMeasuredHit()->
			       GetGede()->GetRMax()),
		      float(recSign*realSign)) ;
    }
  }
#endif // __OTR_USE_ROOT__
//   HitContainer::iterator hitIt= GetHits().begin() ;
  //   ArtePointer<MTRA> mtra((ArtePointer<MIMP>(ArteRelationIterator<MIMP>((*hitIt)->GetMeasuredHit()->GetHITB()->mimp)))->mtra)   ;
//   for(; hitIt!=GetHits().end() ; hitIt++) 
//     if(mtra!=
//        ArtePointer<MTRA>((ArtePointer<MIMP>(ArteRelationIterator<MIMP>((*hitIt)->GetMeasuredHit()->GetHITB()->mimp)))->mtra)) break ;
//   if(hitIt!=GetHits().end()) return false ;
  
  return success ;
} 

bool
poirotCluster::ValidNHits(const poirotReconstructedHit* doNotTakeThisHit) const
{
  //cout << "poirotTrack::ValidNHits()" << endl ;
  float slId[] = {0,0,0} ;
  for(HitContainer::const_iterator hitIt=GetHits().begin() ;
      hitIt!=GetHits().end(); hitIt++) 
    if(*hitIt!=doNotTakeThisHit)
      slId[(*hitIt)->GetMeasuredHit()->GetGede()->GetEStereoAngle()]++;

  return (slId[0]>0) && (slId[1]>0) && (slId[2]>0) ; 
};


void
poirotCluster::Dump() const
{
  cout << "poirotCluster::Dump()" << endl ; 
  for(ClusterContainer::const_iterator zeroIt=fSectorClusters.begin() ;
      zeroIt!=fSectorClusters.end(); zeroIt++) (*zeroIt)->Dump() ;
}

// void 
// poirotCluster::RawWireFit()
// {
//   HepSymMatrix d2Chi2dQ2(2,0) ;
//   HepVector dChi2dQ(2,0) ;
//   HepVector dResdQ(2,0) ;
//   float res ;
//   float weight ;
//   for(ClusterContainer::const_iterator clusIt=fSectorClusters.begin() ;
//       clusIt!=fSectorClusters.end() ; clusIt++) {
//     weight = 12/sqr((*clusIt)->GetGede()->GetWirePitch()) ;
//     res = - (*clusIt)->GetRawX() ;
//     dResdQ[0] = (*clusIt)->GetRawZ()* (*clusIt)->GetCosStereoAngle() ;
//     dResdQ[1] = (*clusIt)->GetRawZ()*-(*clusIt)->GetSinStereoAngle() ;
//     dChi2dQ += 2*res*dResdQ ;
//     d2Chi2dQ2[0][0] += 2*dResdQ[0]*dResdQ[0] ;
//     d2Chi2dQ2[0][1] += 2*dResdQ[0]*dResdQ[1] ;
//     d2Chi2dQ2[1][1] += 2*dResdQ[1]*dResdQ[1] ;
//   }
//   int ierr;
//   HepSymMatrix covOver2 = d2Chi2dQ2.inverse(ierr) ;
//   if(ierr) {
//     cout << "Matrix inversion error in poirotCluster::WireFit()" << endl ;
//     return ;
//   };
//   SetQ(-covOver2*dChi2dQ) ;
//   SetCov(2*covOver2) ;
// }
