#include "poirot/Gede.hh"
#include "poirot/Hit.hh"
#include "poirot/NodeProjection.hh"
#include "poirot/ReconstructedHit.hh"
#include "poirot/TrackDescription.hh"

//__________________________________________________________________
//
// NodeProjection
//
// This class stores some values which have to do with the
// trackfit. It is only meant to improve speed. All its data members
// can be calculated from an HITB or PoirotonstructedHit and a
// corresponding PoirotTrack.
//
// Author: Wouter Hulsbergen 
//__________________________________________________________________


template<class T>
inline T sqr(const T& x) { return x*x; }

inline void
poirotNodeProjection::InitGeometry(const poirotHit* wire)
{
  //double theta = -1*wire->GetGede()->GetStereoAngle() ;
  fZw = wire->GetRawZ() ;
  fProjVector[0] =  wire->GetGede()->GetCosStereoAngle() ; //cos(theta) ;
  fProjVector[1] = -wire->GetGede()->GetSinStereoAngle() ; //sin(theta) ;
  fProjVector[2] = fProjVector[0]*fZw ;
  fProjVector[3] = fProjVector[1]*fZw ;
  fXw = wire->GetRawX() ;
  fSigmaC2 = sqr(wire->GetGede()->GetWirePitch())/12 ;
  fSigmaR2 = sqr(wire->GetDriftRadiusError()) ;
}

inline void
poirotNodeProjection::Init(const poirotHit* measuredhit)
{
  InitGeometry(measuredhit) ;
  fRm = measuredhit->GetDriftRadius() ;
}
  
poirotNodeProjection::poirotNodeProjection(poirotReconstructedHit* hit) 
  : fReconstructedHit(hit), fProjVector(4), fdChi2dQ(), fd2Chi2dQ2()
{
  Init(fReconstructedHit->GetMeasuredHit()) ;
} ;

poirotNodeProjection::poirotNodeProjection(const poirotHit* hit) 
  : fReconstructedHit(0), fProjVector(4), fdChi2dQ(), fd2Chi2dQ2()
{
  Init(hit) ;
} ;

poirotNodeProjection::~poirotNodeProjection() {}

void poirotNodeProjection::Project(const poirotTrackDescription* track,
				bool updateInternal, 
				bool calcDerivatives)
{
  float tprojection = 
    track->GetTx()*fProjVector[0] + track->GetTy()*fProjVector[1] ;
  float noemer2 = 1 + sqr(tprojection) ;
  float noemer = sqrt( noemer2 ) ;
  
  int dim = track->GetDim() ;
  // I have added the z-coordinate where (x,y) is calculated, to simplify
  // the addition of a target constraint different from (0,0,0) ;
  // fD = ((track->GetX()-track->GetTx()*track->GetZ()) *fProjVector[0]+
  // 	(track->GetY()-track->GetTy()*track->GetZ()) *fProjVector[1]+
  // 	track->GetTx()*fProjVector[2]+trackGetTy()*fProjVector[3]
  // 	-fXw) / noemer ;
  
  if(dim==4) fD = ( dot(track->GetQ(),fProjVector) - fXw) / noemer ;
  else fD = ( track->GetTx()*fProjVector[2] + track->GetTy()*fProjVector[3] - fXw) / noemer ; 
  
  if(calcDerivatives) {
    if(fdDdQ.num_row()!=dim) {
      fdDdQ = HepVector(dim) ;
      //fdChi2dQ = HepVector(dim) ;
      fd2Chi2dQ2 = HepSymMatrix(dim) ;
    };
    
    if(dim==4) {
      //fD = ( dot(track->GetQ(),fProjVector) - fXw) / noemer ;
      fdDdQ = fProjVector/noemer ;
      // Are these negligable ?!
      fdDdQ[2] += fD * -fProjVector[0]*tprojection/noemer2 ;
      fdDdQ[3] += fD * -fProjVector[1]*tprojection/noemer2 ;
    } else {
      //fD = ( track->GetTx()*fProjVector[2] + track->GetTy()*fProjVector[3] - fXw) / noemer ; 
      //fdDdQ[0] = fProjVector[2]/noemer + fD * -fProjVector[0]*tprojection/noemer2 ;
      // fdDdQ[1] = fProjVector[3]/noemer + fD * -fProjVector[1]*tprojection/noemer2 ;
      fdDdQ[0] = fProjVector[2]/noemer ;
      fdDdQ[1] = fProjVector[3]/noemer ;
    };
    
    int signD = 1; if(fD<0.) signD=-1;
    int lrsign = signD;
    if(fReconstructedHit) {
      if(updateInternal && !fReconstructedHit->GetLRSignIsFixed()) 
	fReconstructedHit->SetLRSign(fD) ;
      lrsign = fReconstructedHit->GetLRSign() ;
      if(!fReconstructedHit->GetLRSignIsFixed() && lrsign*signD <0) fEtaError = true ;
      else fEtaError = false ;
    };
    
    // Calculate residual
    fResidualR =  (signD*fD) - (lrsign*signD*fRm) ;
    fResidualD = fD - lrsign*fRm ;
    
    float weight = 1/fSigmaR2 ;
    if(lrsign==0) weight = 1/fSigmaC2 ;
    
    fChi2 = weight*sqr(fResidualD) ;
    fdChi2dQ = 2*weight*fResidualD * fdDdQ ; //fResidualR*signD  
    for(int row=0; row<dim; row++)
      for(int col=0; col<=row; col++)
	fd2Chi2dQ2[row][col] = 2*weight * fdDdQ[row]*fdDdQ[col] ;
  }
}

void
poirotNodeProjection::Dump() const 
{
  cout << "fD: " << fD << " fRm: " << fRm << endl
       << "fResidual: " << fResidualR << endl 
       << "fSigma2: " << fSigmaR2 << endl 
       << "fChi2: " << fChi2 << endl
       << "fdChi2dQ: " << fdChi2dQ
       << "fdDdQ: " << fdDdQ << endl ;
}

