#include <arte/arte.hh>
#include <arte/artefunc.h>
#include <algorithm>
#include "poirot/Reco.hh"
#include "poirot/Track.hh"
#include "poirot/TrackCandidate.hh"
#include "poirot/Cluster.hh"

#ifdef __OTR_USE_ROOT__
#include <TH2.h>
#endif

//__________________________________________________________________
// poirotReco
//
// Container for:
// * SuperLayerRecos
// Functionality:
// * Reconstruction of tracks
//
// Author: Wouter Hulsbergen && Ivar Siccama
//__________________________________________________________________

poirotReco* poirotReco::fgMe=0 ;
bool poirotReco::fgAppendClustersToRSEG = false ;

//#define __OTR_RECOINFO__

poirotReco*
poirotReco::Create()
{
  if(!fgMe) fgMe = new poirotReco() ;
  return fgMe ;
}

void
poirotReco::Delete()
{
  delete fgMe ;
  fgMe=0 ;
}

poirotReco::~poirotReco() {
  // Destructor. 
for( TrackContainer::iterator tr=fTrackCandidates.begin();
       tr != fTrackCandidates.end(); tr++) delete (*tr);
  if(fgMe==this) fgMe=0;

}

poirotReco::poirotReco() 
{
  cout << "poirotReco::poirotReco()" << endl ;
}

void
poirotReco::Reset()
{
  //cout << "poirotReco::Reset()" << endl ;
  // Delete tracks
  for(TrackContainer::iterator tr=fTrackCandidates.begin();
      tr != fTrackCandidates.end(); tr++) delete (*tr);
  fTrackCandidates.erase( fTrackCandidates.begin(), fTrackCandidates.end() );

  // Call Reset function for SuperLayerRecos
  for(SuperLayerRecoContainer::iterator slEventIt = fSuperLayerRecos.begin();
      slEventIt != fSuperLayerRecos.end(); slEventIt++) 
    (*slEventIt).second.Reset() ;
}

void
poirotReco::ProcessOneEvent() 
{
  //cout << "poirotReco::ProcessOneEvent()" << endl ;
  // Reset everything
  Reset() ;
  
  for(ArteTable<HITB>::iterator hitIt = ArteTable<HITB>::begin() ;
      hitIt != ArteTable<HITB>::end() ; hitIt++) {
    poirotHit* otrhitb = poirotHit::Create(*hitIt) ;
    if(otrhitb) {
      fSuperLayerRecos[otrhitb->GetGede()->GetSuperLayerName()].Add(otrhitb) ;
    }
    else delete otrhitb ;
  }    
  // Reconstruct clusters
  SearchClusters() ;
  AppendClustersToRSEG() ;

  // Reconstruct tracks
  //SearchTracks() ;

}

int
poirotReco::SearchClusters() 
{
  //cout << "poirotReco::SearchClusters()" << endl ;
  // Search clusters in all superlayers.
  int n=0;
  for(SuperLayerRecoContainer::iterator slEventIt = fSuperLayerRecos.begin();
      slEventIt != fSuperLayerRecos.end(); slEventIt++)
    //if((*slEventIt).first==2) // only reconstruct MC2 (need some speed :)
      n += (*slEventIt).second.SearchClusters() ;
  return n;
}

int
poirotReco::GetNClusters() const 
{
  // Return the number of hits in the event.
  int n=0;
  for(SuperLayerRecoContainer::const_iterator 
	slEventIt = fSuperLayerRecos.begin();
      slEventIt != fSuperLayerRecos.end(); slEventIt++)
    n += (*slEventIt).second.GetClusters().size() ;
  return n;
}

void
poirotReco::DumpHits(class ostream &os) const 
{
  // Call poirotHit::Dump for each hit.
  for(SuperLayerRecoContainer::const_iterator 
	slEventIt = fSuperLayerRecos.begin();
      slEventIt != fSuperLayerRecos.end(); slEventIt++)
    (*slEventIt).second.DumpHits(os) ;
}

void
poirotReco::DumpClusters(class ostream &os) const 
{
  // Call poirotCluster::Dump for each cluster.
  for(SuperLayerRecoContainer::const_iterator slEventIt = fSuperLayerRecos.begin();
      slEventIt != fSuperLayerRecos.end(); slEventIt++)
    (*slEventIt).second.DumpClusters(os) ;
}

// void
// poirotReco::SortHits() 
// {
//   // Call poirotSuperLayerReco::SortHits for each superlayer, which sort
//   // hits in gede and wirenumber.
//   for(SuperLayerRecoContainer::iterator slEventIt = fSuperLayerRecos.begin();
//       slEventIt != fSuperLayerRecos.end(); slEventIt++)
//     (*slEventIt).second.SortHits();
// }

inline bool CompareTargetChi2( poirotTrackCandidate* lhs, 
			       poirotTrackCandidate* rhs) {
  return (lhs->GetClusters().size()> 
	  rhs->GetClusters().size()) ||
    ((lhs->GetClusters().size()==
      rhs->GetClusters().size()) && 
     lhs->GetTargetChi2() < rhs->GetTargetChi2()) ;
}

int
poirotReco::SearchTracks() 
{
  // The track search algorithm defines tarcks in the following way:
  // 1) A track is made up of hit clusters 
  // 2) A track has no cluster or one cluster per superlayer 
  // 3) No track can be `sub-track' of another track
  // It combines clusters in different superlayers into tracks.

  AppendClustersToRSEG() ;

#ifdef __OTR_USE_ROOT__
  static TH2F* rawTrackXYH2 = new TH2F("rawTrackXYH2","",100,-100,100,100,-100,100) ;
  static TH2F* tcTxVsTxH2 = 
    new TH2F("tcTxVsTxH2","",100,-0.26,0.06,100,-0.14,-0.04) ;
  //new TH2F("tcTxVsTxH2","",100,-0.3,0.06,100,-0.14,-0.0) ;
  static TH2F* tcTxVsTxCutH2 = 
    new TH2F("tcTxVsTxCutH2","",100,-0.26,0.06,100,-0.14,-0.04) ;
  //new TH2F("tcTxVsTxH2","",100,-0.3,0.06,100,-0.14,-0.0) ;
#endif
  if(fSuperLayerRecos.size()<2) { return 0 ; } ;

  // ******** Delete previous *********************
  for(TrackContainer::iterator trackIt=fTrackCandidates.begin() ;
      trackIt != fTrackCandidates.end() ; trackIt++)
    delete *trackIt ;
  fTrackCandidates.erase(fTrackCandidates.begin(),fTrackCandidates.end()) ;
  TrackContainer trackSeeds ;

  /** In the following the difference between a seed and a candidate
      is quite arbitrary: I just need two containers. **/
  //for(SuperLayerRecoContainer::reverse_iterator sleIt=fSuperLayerRecos.rbegin() ;
  //    sleIt!=fSuperLayerRecos.rend(); sleIt++) {
  for(SuperLayerRecoContainer::iterator sleIt=fSuperLayerRecos.begin() ;
      sleIt!=fSuperLayerRecos.end(); sleIt++) {
    vector<bool> clusterIsAdded((*sleIt).second.GetClusters().size(),false) ;
    /*** Loop over all seedtracks, combine with clusters and if valid we
         have new candidate ***/
    for(TrackContainer::iterator trackIt=fTrackCandidates.begin() ;
	trackIt!=fTrackCandidates.end() ; trackIt++) {
      poirotTrackCandidate* trackSeed = new poirotTrackCandidate(**trackIt) ;
      bool clusterFound=false ;
      int clId=0;
      for(poirotSuperLayerReco::ClusterContainer::const_iterator clIt=
	    (*sleIt).second.GetClusters().begin() ;
	  clIt!=(*sleIt).second.GetClusters().end() ; clIt++, clId++) {
	if(trackSeed->AddIfValid(*clIt)) {
	  clusterIsAdded[clId] = true ;
	  clusterFound=true ;
	  trackSeeds.push_back(trackSeed) ;
	  trackSeed = new poirotTrackCandidate(**trackIt) ;
	} ;
#ifdef __OTR_USE_ROOT__
	tcTxVsTxH2->Fill(float((*clIt)->GetTx()), 
			 float(trackSeed->GetClusters().front()->GetTx())) ;
	if(fabs((*clIt)->GetTy()-trackSeed->GetClusters().front()->GetTy())<0.05)
	  tcTxVsTxCutH2->Fill(float((*clIt)->GetTx()), 
			      float(trackSeed->GetClusters().front()->GetTx())) ;
#endif //__OTR_USE_ROOT__
      }
      /** If no corresponding cluster is found, the original track is
          added to the list of seeds **/
      if(!clusterFound) trackSeeds.push_back(trackSeed) ;
      else delete trackSeed ;
    }
    /** If there is no track found with a cluster, the cluster itself
        is a new trackcandidate **/
    int clId=0;
    for(poirotSuperLayerReco::ClusterContainer::const_iterator clIt=
	  (*sleIt).second.GetClusters().begin() ;
	clIt!=(*sleIt).second.GetClusters().end() ; clIt++, clId++)
      if(!clusterIsAdded[clId])
	trackSeeds.push_back(new poirotTrackCandidate(*clIt)) ;
    /** All seed track become candidates, and all previous candidates
        are deleted **/
    fTrackCandidates.swap(trackSeeds) ;
    for(TrackContainer::iterator trackIt=trackSeeds.begin() ;
	trackIt != trackSeeds.end() ; trackIt++)
      delete *trackIt ;
    trackSeeds.erase(trackSeeds.begin(),trackSeeds.end()) ;
  }
  /*** (Remove all track candidates which have less than 2 clusters
       assigned to them) ***/
  for(TrackContainer::iterator trackIt=fTrackCandidates.begin() ;
      trackIt!=fTrackCandidates.end() ; trackIt++)
    if((*trackIt)->GetClusters().size()<2) {
      delete *trackIt ;
      fTrackCandidates.erase(trackIt--) ;
    } 
#ifdef __OTR_RECOINFO__
  cout << "Found " << fTrackCandidates.size() << " track candidates " << endl ;
#endif

  /*** Fit all track candidates in global trackfit and remove if fit fails ***/
  for(TrackContainer::iterator trackIt=fTrackCandidates.begin() ;
      trackIt!=fTrackCandidates.end() ; trackIt++)
    if(!((*trackIt)->RangerFit())) {  //if(!((*trackIt)->Fit(true))) {
      delete *trackIt ;
      fTrackCandidates.erase(trackIt--) ;
    } ;
#ifdef __OTR_RECOINFO__
  cout << "Left " << fTrackCandidates.size() << " after track fit . " << endl ;
#endif

  /*** Order tracks in `target compatibility' ***/
  //fTrackCandidates.sort(CompareTargetChi2) ;
  sort(fTrackCandidates.begin(),fTrackCandidates.end(),CompareTargetChi2) ;
  
  /*** Clean up overlapping tracks ***/
  for(TrackContainer::iterator trackIt=fTrackCandidates.begin() ;
      trackIt!=fTrackCandidates.end() ; trackIt++) {
    TrackContainer::iterator trackIt2 = trackIt;
    trackIt2++;
    for( ; trackIt2!=fTrackCandidates.end() ; trackIt2++) {
      if((*trackIt2)->CleanUp(*(*trackIt)) && !((*trackIt2)->Valid())) {
	delete *trackIt2 ;
	fTrackCandidates.erase( trackIt2-- ) ;
      } 
    }
  }
#ifdef __OTR_RECOINFO__
  cout << "and " << fTrackCandidates.size() << " after clean up." << endl ;
#endif  
  return fTrackCandidates.size() ;
}
  
// int
// OTrReco::FitTracks()
// {  
//   int n=0;
//   for(TrackContainer::iterator trackIt = fTrackCandidates.begin() ;
//       trackIt != fTrackCandidates.end() ; trackIt++)
//     //  (*trackIt).RangerFit() ;
//     if((*trackIt)->Fit()) {
//       n++;
//       //(*trackIt)->RangerFitWouterTrack() ;
//     };
//   return n ;
// }

void
poirotReco::AppendRSEG()  
{
  // Copy reconstructed tracks into the RSEG bank. If
  // fgAppendClustersToRSEG is set, not the tracks, but the clusters
  // are appended to RSEG.
  if(!fgAppendClustersToRSEG) {
    for(TrackContainer::iterator trackIt = fTrackCandidates.begin() ;
	trackIt != fTrackCandidates.end() ; trackIt++)
      (*trackIt)->AppendRSEG();
  } else {
    AppendClustersToRSEG() ;
  }
}

void
poirotReco::AppendClustersToRSEG()
{
  for(SuperLayerRecoContainer::iterator slEventIt = fSuperLayerRecos.begin();
      slEventIt != fSuperLayerRecos.end(); slEventIt++)
    (*slEventIt).second.AppendRSEG() ;
}
