// ----------------------------------------------------------------------
//
// Fills the GTAR bank from the DTAR bank.
//
// The DTAR bank must have to be filled with a TGTS_RAW structure 
// on the SLT/TLT farm.
//
//
// TGTS_RAW structure (-> include/tgts_raw.h in the ONLINE library):
// ------------------
//
//    #define TGTS_Nwires 8
//    struct TGTS_RAW {
//      int time; /* somehow mizzi is using int for timestamps, not long */
//      float ia;
//      float bg;
//      float pos[TGTS_Nwires];
//    };
//    typedef struct TGTS_RAW TGTS_RAW;
//
//
// DTAR format:   1.     row:   recid
// ------------   2./ 3. row:   SLB_SUBHDR
//                4.-14. row:   TGTS_RAW
//
//
// Its assumed that the order in the DTAR bank is the same as in the
// target database:
//
//                                                              
// 1999 data:  1. GTAR row  Ia    Al           z=-1.0 cm 
// ----------  2.      row  Ib    Ti           z=-1.0 cm 
//             3.      row  Ii    Al           z=-0.7 cm 
//             4.      row  Io    Al           z=-0.7 cm 
//             5.      row IIa    Fe  (45 deg) z=-5.0 cm 
//             6.      row IIb    Al  (45 deg) z=-5.0 cm 
//             7.      row IIi    C            z=-4.7 cm 
//             8.      row IIo    Ti           z=-4.7 cm 
//
// exp02 data: 1. GTAR row  Ia    Ti           z=-1.3 cm 
// ----------  2.      row  Ib    Ti           z=-1.4 cm 
//             3.      row  Ii    Ti           z=-0.9 cm 
//             4.      row  Io    Ti           z=-0.9 cm 
//             5.      row IIa    Al           z=-5.5 cm 
//             6.      row IIb    W            z=-5.5 cm 
//             7.      row IIi    C            z=-4.9 cm 
//             8.      row IIo    Ti           z=-4.8 cm 
//
// Dimensions for all targets:
//          Length   : 3.0   cm
//          Width    : 0.005 cm =  50 um
//          Thickness: 0.05  cm = 500 um
//
//
// Author: Rainer Wanke, 22-Feb-1999
//  Modification:
//   -- dropping of GTAR is now included in rc99AStandAlone
//                                               01/07/99   Samtleben
//   -- new z positions for exp02 added
//   -- alignment setup from environment variables
//                                               06/09/99   T.Oest
//----------------------------------------------------------------------

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>

// Arte banks,  C++ version
#include "arte/arte.hh"
//extern "C" void mmdrop_(const int &);

// target structure in DTAR bank
#include "tgts_raw.h"

// event structures
#include "event.h"

// byte swapping
#include "esp.h"

// return codes
#include "grr.h"


int rc99AfillGTARfromDTAR()
{
  // Do nothing for MC events
  if (ArteTable<MTRA>::begin() != ArteTable<MTRA>::end()) return GRR_OK;
  
  // print out DTAR record id
  ArtePointer<DTAR> dtar(1);

  // check correct size of DTAR bank
  int dtar_size = ArteTable<DTAR>::size() * sizeof(DTAR);
  if ( dtar_size != sizeof(int) + sizeof(SLB_SUBHDR) + sizeof(TGTS_RAW) ) {
    cout << "rc99AfillGTARfromDTAR: ERROR: DTAR size = " << dtar_size
	 << ", should be " << sizeof(int) + sizeof(TGTS_RAW) << endl;
    cout << "                  No GTAR update done !" << endl;
    return GRR_FISHY;
  }

  // set up target structure from DTAR bank
  dtar += sizeof(SLB_SUBHDR)/sizeof(int) + 1;
  TGTS_RAW *target = (TGTS_RAW*) dtar->bits;
   

//------------------------------------------------------------------------
// set rotation matrix and z positions
//------------------------------------------------------------------------

  float rote[9] = { 1.0, 0.0, 0.0,    // unit matrix
		    0.0, 1.0, 0.0,
		    0.0, 0.0, 1.0};
  float rot[9];       // rotation matrix for 45 degree wires 
  rot[0] =  sqrt(0.5);
  rot[1] =  sqrt(0.5);
  rot[2] =  0.0;
  rot[3] = -sqrt(0.5);
  rot[4] =  sqrt(0.5);
  rot[5] =  0.0;
  rot[6] =  0.0;
  rot[7] =  0.0;
  rot[8] =  1.0;


  float zWirePosition[8];
  if (ArtePointer<EVHD>(1)->exp == 2){
   zWirePosition[0] = -1.3;
   zWirePosition[1] = -1.4;
   zWirePosition[2] = -0.9;
   zWirePosition[3] = -0.9;
   zWirePosition[4] = -5.5;
   zWirePosition[5] = -5.2;
   zWirePosition[6] = -4.9;
   zWirePosition[7] = -4.8;
  } else {
   zWirePosition[0] = -1.0;
   zWirePosition[1] = -1.0;
   zWirePosition[2] = -0.7;
   zWirePosition[3] = -0.7;
   zWirePosition[4] = -5.0;
   zWirePosition[5] = -5.0;
   zWirePosition[6] = -4.7;
   zWirePosition[7] = -4.7;
  };

//------------------------------------------------------------------------
// get the alignment offset from environment variables
//------------------------------------------------------------------------

  float Offset[8] = {0,0,0,0,0,0,0,0};
  char  cOffset[20];
  char *envWireOffset;

  cOffset[0]    = 0;
  envWireOffset = getenv("WIRE_OFFSET_A1");
  if (envWireOffset) {
    strcat(cOffset,envWireOffset);
    Offset[0] = float(atof(cOffset));
  };

  cOffset[0]    = 0;
  envWireOffset = getenv("WIRE_OFFSET_B1");
  if (envWireOffset) {
    strcat(cOffset,envWireOffset);
    Offset[1] = float(atof(cOffset));
  };

  cOffset[0]    = 0;
  envWireOffset = getenv("WIRE_OFFSET_I1");
  if (envWireOffset) {
    strcat(cOffset,envWireOffset);
    Offset[2] = float(atof(cOffset));
  };

  cOffset[0]    = 0;
  envWireOffset = getenv("WIRE_OFFSET_O1");
  if (envWireOffset) {
    strcat(cOffset,envWireOffset);
    Offset[3] = float(atof(cOffset));
  };

  cOffset[0]    = 0;
  envWireOffset = getenv("WIRE_OFFSET_A2");
  if (envWireOffset) {
    strcat(cOffset,envWireOffset);
    Offset[4] = float(atof(cOffset));
  };

  cOffset[0]    = 0;
  envWireOffset = getenv("WIRE_OFFSET_B2");
  if (envWireOffset) {
    strcat(cOffset,envWireOffset);
    Offset[5] = float(atof(cOffset));
  };

  cOffset[0]    = 0;
  envWireOffset = getenv("WIRE_OFFSET_I2");
  if (envWireOffset) {
    strcat(cOffset,envWireOffset);
    Offset[6] = float(atof(cOffset));
  };

  cOffset[0]    = 0;
  envWireOffset = getenv("WIRE_OFFSET_O2");
  if (envWireOffset) {
    strcat(cOffset,envWireOffset);
    Offset[7] = float(atof(cOffset));
  };


//------------------------------------------------------------------------
// Now fill the GTAR bank
//------------------------------------------------------------------------

  GTAR gtar;
  gtar.cmp   = 0;
  gtar.x0bm  = 0.;
  gtar.y0bm  = 0.;
  gtar.sxbm  = 0.;
  gtar.sybm  = 0.;
  gtar.shap  = 1;
  gtar.ms    = 0.;

  // loop over wires
  for (int i=0; i<TGTS_Nwires; i++) {

    // convert wire position to cm and correct sign if necessary
    float WirePosition[TGTS_Nwires];
    ESP_get_int( (int*) target->pos, (int*) WirePosition, TGTS_Nwires );
    if ( i==1 || i==5 || i==3 || i==7 )
      WirePosition[i] =  0.1 * WirePosition[i];
    else
      WirePosition[i] = -0.1 * WirePosition[i];


    // wire number and thickness
    gtar.wnr = i+1;
    gtar.dz  = 0.05;

    // I above/below
    if ( i==0 || i== 1 ) {
      gtar.z   = zWirePosition[i];
      gtar.dx  = 3.0;
      gtar.dy  = 0.005;
      gtar.x   = 0.;
      gtar.y   = WirePosition[i] + Offset[i];
      gtar.rot = rote;
    }

    // I inner/outer
    if ( i==2 || i== 3 ) {
      gtar.z   = zWirePosition[i];
      gtar.dx  = 0.005;
      gtar.dy  = 3.0;
      gtar.x   = WirePosition[i] + Offset[i];
      gtar.y   = 0.;
      gtar.rot = rote;
    }

    // II above/below  (rotated by 45 degrees for exp01!)
    if ( i==4 || i== 5 ) {
      gtar.z   = zWirePosition[i];
      gtar.dx  = 3.0;
      gtar.dy  = 0.005;
      gtar.x   = 0.;
      gtar.y   = WirePosition[i] + Offset[i];
      if (ArtePointer<EVHD>(1)->exp == 2) {
       gtar.rot = rote;
      } else {
       gtar.rot = rot;
      };
    }

    // II inner/outer
    if ( i==6 || i== 7 ) {
      gtar.z   = zWirePosition[i];
      gtar.dx  = 0.005;
      gtar.dy  = 3.0;
      gtar.x   = WirePosition[i] + Offset[i];
      gtar.y   = 0.;
      gtar.rot = rote;
    }

    // Store new GTAR table
    ArteTable<GTAR>::push_back( gtar );

  }

  return GRR_OK;
}




