//-----------------------------------------------------------------------//
//
//  RingSearchARTE: ARTE interface for stand-alone ring reconstruction
//
//  Author: Reinhard Eckmann
//
//  Created: May-98
//
//-----------------------------------------------------------------------//

#include "rise/RingSearchARTE.h"
#include "rich/rich.h"
#include "arte/RSEG_cmp.hh"
#include "CLHEP/Hist/HBookHistogram.h"
#include "rise/RICHFit.h"

// 
// --- With RISE_TIMING defined, a root file rise_timing.root will be
// --- created in working directory containing a 2d histogram:
// --- photon occupancy vs. time for RingSearch execution per event
//
//#define RISE_TIMING

RingSearchARTE::RingSearchARTE() : RingSearchDB(), RingSearchTiming()  {

  // -- maybe this was already filled by ricini?
  if (ArteTable<MTRA>::nrow()) {

    for (unsigned i=1;i<=ArteTable<GESL>::nrow();i++) {
      ArtePointer<GESL> gesl(i);
      if (gesl->cmp==12) {
	RichRun::magnet=gesl->mfl;
	break;
      }
    }
 
    for (ArteTable<GEDE>::iterator igede=ArteTable<GEDE>::begin();
       igede!=ArteTable<GEDE>::end(); ++igede) {
      if ((*igede)->cmp==5) {
	RichRun::gede=(*igede);
	break;
      }
    }

  }

  if (RichRun::magnet) printf("RingSearch with magnet ON\n");
  else printf("RingSearch with magnet OFF\n");
}


void RingSearchARTE::makeHitMap() {

  h.reset();

  int plane,imtra;

  int n0=RichRun::gede->hit1.index();
  int n1=RichRun::gede->hitl.index();

  if (n1-n0>OccupancyCut) return;

  int i;
  for (i=n0;i<n1;i++) {

    ArtePointer<HITB> hitb(i);
    ArteRelationIterator<DIGB> idigb=hitb->digb.begin();
    ArtePointer<DIGB> digb(idigb);
    int id=digb->id;
    plane = (smFromChannelID(id)>6) ? 1 : 0 ;
    if (ArtePointer<MTRA>::nrow()) { // MC mode
      ArteRelationIterator<MIMP> imimp=hitb->mimp;
      ArtePointer<MTRA> mtra=(*imimp)->mtra;
      imtra = mtra.index();
    }
    else { // DAQ mode
      imtra=99999;
    }
    h.append(d.dualx(hitb->q*1000),d.dualy(hitb->r*1000),plane,imtra);
  }
  h.randomize(); 
  d.nextevent();
}


void RingSearchARTE::getrings(float r1,float r2,float dr,int t) {
#ifdef RISE_ROOT
#ifdef RISE_TIMING
  startTiming();
#endif
#endif
  makeHitMap();
  if (dr<0)
    for (float radius=r2;radius>r1;radius+=dr) {// mrad /////
      d.define(radius,t);
      d.ringsearch(h);
    }
  else
    for (float radius=r1;radius<r2;radius+=dr) {// mrad /////
      d.define(radius,t);
      d.ringsearch(h);
    }
  reFit();
  FillRSEG();
#ifdef RISE_ROOT
#ifdef RISE_TIMING
  stopTiming();
  fillTimingHistograms();
#endif
#endif
}


void RingSearchARTE::getrings() {
  if (!ArteTable<MTRA>::nrow()) getSearchParameters(ArtePointer<EVHD>(1)->run,0);
  getrings(minSearchRadius(),maxSearchRadius(),searchStep(),threshold());
}



void RingSearchARTE::FillRSEG() {

  RSEG r;
  ArtePointer<RSEG>rseg;
  //cout<<"RingSearchARTE::FillRSEG() numrings="<<numrings()<<endl;
  for (int i=0;i<numrings();i++) {

    Ring &ring=getRing(i);

    r.xf=ring.phi();
    r.yf=ring.lambda();

    r.time=ring.thetaC();
    r.dtim=cos(getRadius(ArtePointer<EVHD>(1)->run))/cos(r.time);

    r.txf=tx(i); //=ring.tx();
    r.tyf=ty(i); //=ring.ty();

    r.pf=float(ring.numHits());
    r.ze=ring.numhit_refit;
    r.ye=ring.numhit_bgrd;
    r.xe=ring.numhit_3sigma;
    r.zf=ring.f;
    //r.txe=ring.N;
    //r.tye=ring.mu;
    r.pe=ring.sigma_signal;

    r.chi=ring.chi2;

    r.cvf[ 0]=0;
    r.cvf[ 1]=0;
    r.cvf[ 2]=0;
    r.cvf[ 5]=ring.covariance(0,0);
    r.cvf[ 8]=ring.covariance(0,1);
    r.cvf[12]=ring.covariance(0,2);
    r.cvf[ 9]=ring.covariance(1,1);
    r.cvf[13]=ring.covariance(1,2);
    r.cvf[14]=ring.covariance(2,2);
    r.cvf[ 3]=0;
    r.cvf[ 6]=0;
    r.cvf[10]=0;
    r.cvf[ 4]=0;
    r.cvf[ 7]=0;
    r.cvf[11]=0;

    /*cout<<r.cvf[5]<<" "<<r.cvf[8]<<" "<<r.cvf[12]<<endl;
    cout<<r.cvf[8]<<" "<<r.cvf[9]<<" "<<r.cvf[13]<<endl;
    cout<<r.cvf[12]<<" "<<r.cvf[13]<<" "<<r.cvf[14]<<endl<<endl;;*/

    r.cmp = Rsegc::rich;

    if (ring.numhit_refit>50) r.fit=10; // possible gamma conversion
    else r.fit=8;                       // charged track

    rseg=ArteTable<RSEG>::push_back(r);

    if (ArtePointer<MTRA>::nrow())
      createRelation(rseg,ArtePointer<MTRA>(imtra(i)),&RSEG::mtra);
  }
}



void RingSearchARTE::reFit() {
;
}

RingSearchARTE *RICH_Search=0;

// -------------------------------------------------------------------------------------------------- //


void rcrings_(float r1,float r2,float dr,int t) {
  if (!RICH_Search) RICH_Search=new RingSearchARTE;
  RICH_Search->getrings(r1,r2,dr,t);
}


void rcringsdb_() {
  if (!RICH_Search) RICH_Search=new RingSearchARTE;
  RICH_Search->getrings();
}




