#ifndef __WAX_GEDE_HH__
#define __WAX_GEDE_HH__

#ifndef _MATH_H
#include <math.h>
#endif
#ifndef _STDLIB_H
#include <stdlib.h>
#endif

#include <iostream>

#ifdef USE_ARTE
 #ifndef __ARTE_HH__
 #include "arte/arte.hh"
 #endif
 typedef ArtePointer<GEDE> ArtePointerGEDE ;
#else
 #ifndef _GEDE_HH_
 #include "auto/GEDE.h"
 #include "auto/OTMO.h"
 #endif
 typedef GEDE* ArtePointerGEDE ;
#endif

#ifndef __WAX_BASICTYPES_HH__
#include "BasicTypes.hh"
#endif

class waxRTRelation ;
class istream ;
class ostream ;

/**  ARTE GEDE modules (detector parts) for the moment presented by a
     geometrical box **/

class waxGede
{
public:
  // Enumerates and typedefs definitions
  enum EPitch       { kFiveMM=0, kTenMM=1 } ; // Pitch Type 
  enum EStereoAngle { kMinusFiveDegrees=0, kZeroDegrees=1, kPlusFiveDegrees=2 } ; // Stereoangle type
  enum EXSide       { kXSideMinus=0, kXSidePlus=1 } ;                      // Module orientation type 
  //N.B.: EXSide=0 hold for all module on minus X, but e.g. not for MINI-TC2 !!
  
  // Constructors, destructor, assignment operator
  waxGede();                    // default constructor
  waxGede(const waxGede&);      // copy constructor
#if defined(USE_ARTE)
  waxGede(ArtePointerGEDE gede);// constructor from an ARTE GEDE pointer
#endif
  waxGede(istream& is);         // constructor reading from an input filestream
  ~waxGede();
  const waxGede& operator=(const waxGede&) ;
  
  // Setters and Getters
  static int GetOTrComponentId() { return fgOTrComponentId ; }

  const ArtePointerGEDE& GetGEDE() const { return fGEDE ; }
  int GetGEDEIndex() const { return fGEDEIndex ; }
  int GetSuperLayerId() const { return fSuperLayerId ; }
  const string& GetSuperLayerName() const { return fSuperLayerName; }
  EPitch GetEPitch() const { return fEPitch ; }
  EStereoAngle GetEStereoAngle() const { return fEStereoAngle ; }
  EXSide GetXSide() const { return fEXSide ; }
  int GetWireIndexOffset() const { return fWireIndexOffset ; }
  void SetWireIndexOffset(int offset) { fWireIndexOffset = offset ; }

  // The RTRelations ...
  waxRTRelation* GetRTRelation() const { return fRTRelation ; }
  void SetRTRelation(waxRTRelation* rtr) { fRTRelation = rtr ; }

  // The Geometry
  int GetNMonoLayers() const { return GetGEDE()->nsl ; } ;
  unsigned char GetMonoLayer(WireIndex_t wireNumber) const { return wireNumber/32; }
  float GetRawX(WireIndex_t wireNumber=0) const ;
  //float GetRawY(WireIndex_t wireNumber=0) const ;
  float GetRawY1(WireIndex_t wireNumber=0) const ;
  float GetRawY2(WireIndex_t wireNumber=0) const ;
  float GetRawZ(WireIndex_t wireNumber=0) const ;

  bool WithinYAcceptance(const float tx, const float ty) const ;
  //float GetWireLength() const ;
  //float GetCentralX(WireIndex_t wireNumber=0) const ;
  //float GetCentralY(WireIndex_t wireNumber=0) const ;
  float GetCosStereoAngle() const { return GetGEDE()->rot[0] ; }
  float GetSinStereoAngle() const { return GetGEDE()->rot[1] ; }
  float GetWirePitch()  const { return GetGEDE()->pit ; }
  float GetLayerPitch() const { return 2*(GetGEDE()->zfw[1]-GetGEDE()->zfw[0]) ; }
  int GetNWires() const;
  int GetSectorId() const { return GetGEDE()->sec ; }
  int GetLayerId() const  { return GetGEDE()->lay ; }
  int GetModuleId() const { return atoi(GetGEDE()->hwc+6) ; }
  const char* GetHWC() const { return GetGEDE()->hwc ; }

  static SuperLayerId_t GetSuperLayerId(const char* slid) ;
  static string GetSuperLayerName(SuperLayerId_t slId) ;

private:
  // Methods
  void Initialize() ; // determine EStereoAngle, EPitch, EXSide and SuperLayerId
private:
  ArtePointerGEDE fGEDE ; // ArtePointer to a GEDE
  int fGEDEIndex ;        // gede index
  int fSuperLayerId;      // OTR superlayer ID
  string fSuperLayerName;  // SuperLayername (from GESL)
  int fWireIndexOffset ;

  EStereoAngle fEStereoAngle ; // StereoAngle type
  EPitch fEPitch ;             // Pitch type
  EXSide fEXSide ;             // Orientation type
  waxRTRelation* fRTRelation ; // RTRelation for this module
  static int fgOTrComponentId ;
} ;

/***************** inlines **********************************/

ostream& operator << (ostream&, const waxGede& atgede );

inline float waxGede::GetRawX(WireIndex_t wireNumber) const {
  return GetGEDE()->xfw[GetMonoLayer(wireNumber)] + 
    (wireNumber%32)*GetWirePitch() ; 
} 

inline float waxGede::GetRawY1(WireIndex_t wireNumber) const {
  return GetGEDE()->y1w[GetMonoLayer(wireNumber)] ;
}

inline float waxGede::GetRawY2(WireIndex_t wireNumber) const {
  return GetGEDE()->y2w[GetMonoLayer(wireNumber)] ;
}

inline float waxGede::GetRawZ(WireIndex_t wireNumber) const {
  return GetGEDE()->zfw[GetMonoLayer(wireNumber)] ;
} 

#endif
