#ifndef __WAX_PARIZATION_HH__
#define __WAX_PARIZATION_HH__

#ifndef __WAX_FUNCTIONVALUE_HH__
#include "FunctionValue.hh"
#endif
#ifndef __SGI_STL_VECTOR
#include <vector>
#endif
#ifndef _Vector_H_
#include "CLHEP/Matrix/Vector.h"
#endif
#ifndef _SYMMatrix_H_
#include "CLHEP/Matrix/SymMatrix.h"
#endif

/** Base class for parameterization on [xmin,xmax) **/
class TH1 ;

class waxParization
{
public:
  // typedefs, enumerates
  typedef float XValue ;
  typedef float YValue ;
  enum EParType { kSpline=1, kChebychev=2 } ;

  // destructor
  virtual ~waxParization() {}

  // setters and getters
  XValue GetXMin() const { return fXMin ; }
  XValue GetXMax() const { return fXMax ; }
  const HepVector& GetPars() const { return fPars ; }
  const HepSymMatrix& GetCov() const { return fCov ; }
  void SetXMin(XValue x) { fXMin=x; }
  void SetXMax(XValue x) { fXMax=x; }
  void SetPars(const HepVector& pars,const HepSymMatrix& cov=0 ) {
    fPars = pars ; fCov=cov ; }

  // implemented by derived class
  virtual vector<waxFunctionValue> CalcBasicFuncVector(XValue x) const = 0 ;
  virtual HepVector CalcValueVector(XValue x) const = 0 ;
  virtual void ChangeNPars(int n) = 0;
  virtual void Read(XValue xmin,XValue xmax,istream&) = 0 ;
  virtual void Write(ostream&) const = 0;

  // other methods
  int GetNPars() const { return fPars.num_row() ; }
  YValue CalcValue(XValue x) { return dot(CalcValueVector(x),fPars) ;}
  waxFunctionValue CalcFunction(XValue x) ;
  waxFunctionValue CalcInverse(YValue y, YValue precision) ; // inverse function
  static waxParization* Create(XValue xmin, XValue xmax, 
			       int npars=0,YValue y=0) ;
  static waxParization* Create(XValue xmin, XValue xmax, 
			       const HepVector& pars,
			       const HepSymMatrix& cov=0) ;
#ifdef __OTR_USE_ROOT__
  static waxParization* Create(XValue xmin, XValue xmax, int npars,
			       TH1* histo, YValue y_xmin,YValue y_xmax) ;
#endif

protected:
  waxParization(XValue xmin, XValue xmax, int npars) 
    : fXMin(xmin),fXMax(xmax),fPars(npars,0), fCov(npars,0) {}
  waxParization(XValue xmin, XValue xmax, 
		      const HepVector& pars, const HepSymMatrix& cov) 
    : fXMin(xmin),fXMax(xmax),fPars(pars), fCov(cov) {}
private:
  XValue fXMin ;
  XValue fXMax ;
  HepVector fPars ;
  HepSymMatrix fCov ;
};

class waxSplineParization : public waxParization
{
public:

  // constructors, destructors
  waxSplineParization(XValue xmin, XValue xmax,int npars,YValue y) ;
  waxSplineParization(XValue xmin, XValue xmax,
		      const HepVector& pars,const HepSymMatrix& cov) ;
#ifdef __OTR_USE_ROOT__
  waxSplineParization(XValue xmin, XValue xmax, int npars,
		      TH1* histo, YValue y_xmin, YValue y_xmax) ;
#endif
  // setters and getters
  int GetDegree() { return fDegree ; }
  HepVector GetKnots() { return fKnots ; }

  // declared by base class
  virtual vector<waxFunctionValue> CalcBasicFuncVector(XValue x) const ;
  virtual HepVector CalcValueVector(XValue x) const ;
  virtual void ChangeNPars(int n) ;
  virtual void Read(XValue xmin, XValue xmax,istream&) ;
  virtual void Write(ostream&) const ;

private:
  static HepVector CalcEquidistantKnots(int degree, int npar, 
					XValue xmin,XValue xmax) ;
  static double EvaluateBSpline(const HepVector&,
				const int, const int, const double) ;
  static double EvaluateBSplineDerivative(const HepVector&,
					  const int, const int, const double) ;
  static double EvaluateBSplineSecondDerivative(const HepVector&,
						const int, const int, const double) ;
  int       fDegree ;
  HepVector fKnots ; //!
};

#endif //__WAX_PARIZATION_HH__
