#include <fstream.h>
#include <map>
#include <string>
#include "wax/GedeManager.hh"
#include "wax/Gede.hh"
#include "wax/Hit.hh"

//__________________________________________________________________
// waxGedeManager
//
// Container for:
// * all waxGedes
// Functionality:
// * create waxGedes from file or Arte tables
// * fast decoding from gedeindex to gede pointer
//
// Author: Wouter Hulsbergen && Ivar Siccama
//__________________________________________________________________

waxGedeManager::~waxGedeManager() 
{
  for(GedeContainer::iterator gedeIt=fGedes.begin();
      gedeIt!=fGedes.end() ; gedeIt++) delete *gedeIt ;
}

waxGedeManager::waxGedeManager( const char* configdirname ) 
{ 
  cout << "Building waxGedes from file." << endl;
  string dir(configdirname) ;
  const char* filename = const_cast<char*>((dir + "/caldata/geometry/geometry.txt").c_str()) ;
  char header[5];
  ifstream infile( filename );
  if ( infile.is_open() <= 0 ) {
    cout << "waxDetector: input file " << filename 
	 << " does not exist." << endl;
  } else {
    while ( infile ) {
      infile >> header;
      //cout << "header: " << header << endl;
      if ( !infile ) { break; }
      waxGede* gede = new waxGede( infile );
      if ( !infile ) break;
      if (gede->GetNMonoLayers()>0)
	fGedes.push_back(gede) ;
    }
    infile.close();
  }
  Init() ;
}

#ifdef USE_ARTE
waxGedeManager::waxGedeManager()
{
  cout << "Building waxGedes from the ARTE geometry." << endl;

  for ( ArteTable<GEDE>::iterator itgede = ArteTable<GEDE>::begin() ;
	itgede != ArteTable<GEDE>::end(); itgede++ )
    if ( itgede->cmp == waxGede::GetOTrComponentId() ) {
      waxGede* gede = new waxGede( ArtePointer<GEDE>(itgede) );
      //if (gede->GetNMonoLayers()>0)
      fGedes.push_back(gede) ;
      //else delete gede ;
    }
  Init() ;
}
#else
waxGedeManager::waxGedeManager() {}
#endif

void waxGedeManager::Init()
{
  waxHit::SetGedeManager(this) ;
  Map() ;
  SetWireIndexOffsets() ;
} 

void waxGedeManager::Map()
{
  if(fGedes.empty()) {
    fMinId =0;
  } else {
    GedeContainer::iterator gedeIt=fGedes.begin() ;
    GedeIndex_t minId = (*gedeIt)->GetGEDEIndex() ;
    GedeIndex_t maxId = minId ;
    for(; gedeIt!=fGedes.end() ; gedeIt++) {
      GedeIndex_t id = (*gedeIt)->GetGEDEIndex() ;
      if(minId>id) minId=id;
      else if(maxId<id) maxId=id;
    }
    GedeContainer map(maxId-minId+1,NULL) ;
    for(gedeIt=fGedes.begin();gedeIt!=fGedes.end() ; gedeIt++)
      map[(*gedeIt)->GetGEDEIndex()-minId] = *gedeIt ;
    fGedes = map ;
    fMinId = minId ;
    cout << "waxGedeManager::Map() first/last gede " 
	 << fMinId << " / " << maxId << endl ;
  }

  fHWCMap.erase(fHWCMap.begin(),fHWCMap.end());
  for(GedeContainer::iterator gedeIt=fGedes.begin() ;
      gedeIt!=fGedes.end() ; gedeIt++) 
    if(*gedeIt)
      fHWCMap[string((*gedeIt)->GetGEDE()->hwc)]= *gedeIt ;
}

waxGede* waxGedeManager::GetGede(const string& hwc) const 
{
  waxGede* gede=0;
  HWCMap::const_iterator gedeIt = fHWCMap.find(hwc) ;
  if(gedeIt != fHWCMap.end()) gede = gedeIt->second ;
  return gede ;
}

void waxGedeManager::SetWireIndexOffsets()
{
//   typedef waxGede* Module ;
//   typedef map<int, Module, less<int> > Sector ;
//   typedef map<int, Sector, less<int> > Layer ;
//   typedef map<int, Layer, less<int> >  SuperLayer ;
//   typedef map<int, SuperLayer, less<int> > Detector ;

//   Detector detector ;
//   for(GedeContainer::iterator gedeIt=fGedes.begin() ;
//       gedeIt != fGedes.end() ; gedeIt++)
//     if(*gedeIt) {
//       int slid  = (*gedeIt)->GetSuperLayerId() ;
//       int layid = (*gedeIt)->GetLayerId() ;
//       int secid = (*gedeIt)->GetSectorId() ;
//       int modid = (*gedeIt)->GetModuleId() ;
//       detector[slid][layid][secid][modid] = *gedeIt ;
//     }
  
//   for(Detector::iterator slIt=detector.begin() ;
//       slIt!=detector.end() ; slIt++) {
//     //cout << "SuperLayer: " << slIt->first << endl ;
//     for(SuperLayer::iterator layIt=slIt->second.begin() ;
// 	layIt !=slIt->second.end() ; layIt++) {
//       //cout << "  Layer: " << layIt->first << endl ;
//       for(Layer::iterator secIt=layIt->second.begin() ;
// 	  secIt !=layIt->second.end() ; secIt++) {
// 	//cout << "    Sector: " << secIt->first << endl ;
// 	// is sorted in moduleId then also in x ??
// 	int offset=0;
// 	for(Sector::iterator modIt=secIt->second.begin() ;
//  	    modIt !=secIt->second.end() ; modIt++) 
//  	  {
//  	    modIt->second->SetWireIndexOffset(offset) ;
//  	    offset+=modIt->second->GetNWires() ;
// 	    // 	    cout << "      Module: " << modIt->first 
// 	    // 		 << " Gede: " << modIt->second->GetGEDEIndex() 
// 	    // 		 << " Offset: " << modIt->second->GetPRWireIndexOffset() 
// 	    // 		 << " " <<  modIt->second->GetPRWireIndex(0) 
// 	    // 		 << " X: " << modIt->second->GetRawX() << endl ;
// 	  }
//       }
//     }
//   }
}

#ifdef USE_ARTE
void waxGedeManager::ResetHITBPointers()
{
  for(GedeContainer::iterator gedeIt = fGedes.begin() ;
      gedeIt != fGedes.end() ; gedeIt++) 
    if(*gedeIt) {
      (*gedeIt)->GetGEDE()->hit1 = ArtePointer<HITB>(0)  ;
      (*gedeIt)->GetGEDE()->hitl = ArtePointer<HITB>(-1) ;
    }
}
#endif
