#include <iostream.h>
#include <iomanip.h>
#include "wax/Parization.hh"

/********** waxParization *******************/


#ifdef __OTR_USE_ROOT__
waxParization* waxParization::Create(XValue xmin, XValue xmax, int npars,
				     TH1* histo, 
				     YValue y_xmin,YValue y_xmax) 
{
  waxParization* par(0);
  par = new waxSplineParization(xmin,xmax,npars,histo,y_xmin,y_xmax) ;
  return par;
}
#endif


waxParization*
waxParization::Create(XValue xmin, XValue xmax, 
		      int npars, YValue y)
{
  waxParization* par(0);
  par = new waxSplineParization(xmin,xmax,npars,y) ;
  return par;
}

waxParization*
waxParization::Create(XValue xmin, XValue xmax,
		      const HepVector& pars, const HepSymMatrix& cov)
{
  waxParization* par(0);
  par = new waxSplineParization(xmin,xmax,pars,cov) ;
  return par;
}

waxFunctionValue
waxParization::CalcFunction(XValue x) 
{
  vector<waxFunctionValue> bfv = CalcBasicFuncVector(x) ;
  waxFunctionValue rv;
  for(int i=0; i<fPars.num_row(); i++) rv += fPars[i]*bfv[i] ;
  return rv ;
}

waxFunctionValue 
waxParization::CalcInverse(YValue y, YValue precision)
{
  // calculate inverse using newton
  XValue x = (GetXMax()-GetXMin())/2 ;
  waxFunctionValue fvinx=CalcFunction(x) ;
  YValue deltay = fvinx.fY-y ;
  while(fabs(deltay)>precision) {
    x -= deltay/fvinx.fdYdX ;
    fvinx=CalcFunction(x) ;
    deltay = fvinx.fY-y ;
  }
  return fvinx ;
} 

/********** waxSplineParization *******************/

void waxSplineParization::Read(XValue xmin, XValue xmax, istream& ifs) 
{
  // This looks very clumsy, but I do not want to read xmin and xmax here ...
  SetXMin(xmin) ;
  SetXMax(xmax) ;
  int npars ;
  ifs >> npars ;
  HepVector pars = HepVector(npars) ;
  for(int row=0; row<npars; row++) ifs >> pars[row] ;
  HepSymMatrix cov = HepSymMatrix(npars) ;
  for(int row=0; row<npars; row++)
    for(int col=0; col<=row; col++) ifs >> cov[row][col] ;
  SetPars(pars,cov) ;
  fDegree = min(GetNPars()-1,3) ;
  fKnots = CalcEquidistantKnots(GetDegree(),GetNPars(),GetXMin(),GetXMax()) ;
}

void waxSplineParization::Write(ostream& ofs) const
{
  ofs  << setw(5) << setprecision(3) << GetNPars() << endl ;
  for(int row=0; row<GetNPars(); row++)
    ofs << setw(15) << setprecision(10) << GetPars()[row] ;
  ofs << endl ;

  for(int row=0; row<GetNPars(); row++) {
    for(int col=0; col<=row; col++)
      ofs << setw(15) << setprecision(10) << GetCov()[row][col] ;
    ofs << endl ;
  }
}

waxSplineParization::waxSplineParization(XValue xmin, XValue xmax, 
					 int npars, YValue y)
  : waxParization(xmin,xmax,npars), fDegree(3)
{
  fDegree = min(GetNPars()-1,3) ;
  fKnots = CalcEquidistantKnots(fDegree, GetNPars(),GetXMin(),GetXMax()) ;
  HepVector pars(GetNPars());
  for(int i=0; i<GetNPars(); i++) pars[i]=y;
  SetPars(pars) ;
}

waxSplineParization::waxSplineParization(XValue xmin, XValue xmax,
					 const HepVector& pars,
					 const HepSymMatrix& cov) 
  : waxParization(xmin,xmax,pars,cov)
{
  fDegree = min(GetNPars()-1,3) ;
  fKnots = CalcEquidistantKnots(fDegree, GetNPars(),GetXMin(),GetXMax()) ;
}

vector<waxFunctionValue> 
waxSplineParization::CalcBasicFuncVector(XValue x) const
{
  int npar = fKnots.num_row()-fDegree-1 ;
  vector<waxFunctionValue> vec(npar) ;
  for(int ipar=0; ipar<npar; ipar++) {
    vec[ipar].fY    = EvaluateBSpline(fKnots, ipar, fDegree, x) ;
    vec[ipar].fdYdX = EvaluateBSplineDerivative(fKnots, ipar, fDegree, x) ;
    //    vec[i].d2dx2 =
    //       EvaluateBSplineSecondDerivative(fKnots, ipar, degree, X) ;
    //vec[i].fX    = x;
  }
  return vec ;
}

HepVector 
waxSplineParization::CalcValueVector(XValue x) const 
{
  HepVector Y(GetNPars()) ;
  for(int ipar=0; ipar<Y.num_row(); ipar++)
    Y[ipar] = EvaluateBSpline(fKnots,ipar,fDegree,x) ;
  return Y;
}

#ifdef __OTR_USE_ROOT__
#include <TH1.h>

waxSplineParization::waxSplineParization(XValue xmin,XValue xmax,int npars,
					 TH1* histo,
					 YValue y_xmin,YValue y_xmax) :
  waxParization(npars,xmin,xmax), fDegree(3)
{
  // This constructor determines a set of parameters by inverting a
  // histogram. It is used to determine an rt-ralation from inverting
  // a tdc-spectrum.

  //Set time boundaries
  double estimated_time_Rmin=y_xmin;
  double estimated_time_Rmax=y_xmax;

  double tMin = estimated_time_Rmin ;
  double tMax = estimated_time_Rmax ;
  int minBin = histo->GetXaxis()->FindBin(tMin) ;
  int maxBin = histo->GetXaxis()->FindBin(tMax) ;
  if(minBin>maxBin) {
    int tmp=minBin; minBin=maxBin; maxBin=tmp ; }
  const int noBins = maxBin-minBin + 1 ;
  
  //Integrate spectrum
  vector<float> integratedspectrum(noBins) ;
  for(int j=0; j<noBins; j++) integratedspectrum[j]=0;
  for(int i=0; i<noBins; i++)
    for(int j=i; j<noBins; j++)
      integratedspectrum[j]+= histo->GetBinContent(i+minBin) ;
  
//   for(int j=0; j<noBins; j++)
//     cout << j << " " << integratedspectrum[j] << endl ;

  double noEntries = integratedspectrum[noBins-1] ;
  // Define knots
  fKnots = CalcEquidistantKnots(GetDegree(),GetNPars(),GetXMin(),GetXMax()) ;

  //Define Schoenberg's knots
  HepVector schoenbergknots(npars) ;
  if(fDegree==0)
    for(int i=0; i< npars; i++)
      schoenbergknots[i] = (0.5*(fKnots[i]+fKnots[i+1])) ;
  else
    for(int i=0; i< npars ; i++) {
      double sum(0.) ;
      for(int j = i+1 ; j < i + fDegree + 1; j++) sum += fKnots[j];
      schoenbergknots[i] = sum/(double)fDegree ;
    }
  
  //Set initial parameters by spec value on Schoenberg's knots
  HepDouble norm = 2./noEntries ; // effectiveCellSize/GetRMax() ; 
  HepVector TParameters(npars) ;
  for(int j=0; j< npars ; j++) {
    TParameters[j] = tMax ;
    for(int i=0; i< noBins ; i++) {
//    cout << i << " " << tMin << " " << tMax << " " 
// 	   << (noEntries-integratedspectrum[i])*norm << " "
// 	   << schoenbergknots[j]+1 << endl ;
      if(((tMax>tMin) && (integratedspectrum[i]*norm>=schoenbergknots[j]+1)) ||
	 ((tMin>tMax) && ((noEntries-integratedspectrum[i])*norm<=schoenbergknots[j]+1)))
	{
	  TParameters[j] = histo->GetBinCenter(i+minBin) ;
	  break ;
	}
    }
  }
  SetPars(TParameters) ;
}
#endif

void
waxSplineParization::ChangeNPars(int n_new)
{
  if(n_new != GetNPars()) {
    int k_new = min(n_new-1,3) ;
    HepVector knots_new = CalcEquidistantKnots(k_new,n_new,GetXMin(),GetXMax()) ;
    
    HepVector schoenbergknots(n_new) ;
    if(k_new==0) for(int i=0; i< n_new; i++)
      schoenbergknots[i] = (0.5*(knots_new[i]+knots_new[i+1])) ;
    else for(int i=0; i< n_new ; i++) {
      double sum(0.) ;
      for(int j = i+1 ; j < i + k_new + 1; j++) sum += knots_new[j];
      schoenbergknots[i] = sum/(double)k_new ;
    }
    HepVector par_new = HepVector(n_new,0) ;
    for(int i=0; i<n_new; i++)
      for(int j=0; j<GetNPars(); j++)
	par_new[i] += GetPars()[j]*EvaluateBSpline(fKnots,j,fDegree,schoenbergknots[i]) ;
    SetPars(par_new) ;
    fDegree = k_new ;
    fKnots = knots_new ;
  }
}

/* These are static functions. They just calculate some general spline
things ...  */

#define FUF 1e-16

HepVector
waxSplineParization::CalcEquidistantKnots(const int k /* degree */, 
					  const int n /* number of parameters */,
					  XValue xmin,
					  XValue xmax)
{
  // Calculates equidistant knots belonging to a spline of
  // degree k, with n parameters, and width boudaries {-1,1].
  // The knots on the boundaries are of multiplicity k.

  int m = n + k + 1; // # knots 
  HepDouble deltaXi = (xmax-xmin)/(m - 2*k - 1) ;
  HepVector knots(m) ;
  int i;
  for(i=0; i<k+1; i++)     knots[i] = xmin ;
  for(i=k+1; i<m-k-1; i++) knots[i] = xmin + (i - k) * deltaXi ;
  for(i=m-k-1; i<m; i++)   knots[i] = xmax ;
  return knots ;
}

//Evaluate j'th basis spline of degree k in x on knots t   

double
waxSplineParization::EvaluateBSpline(const HepVector& t,
				     const int j, const int k, 
				     const double x)
{
  //cout << "SplineRTRelation::EvaluateBSpline()" << endl;
  if(j < 0 || j+k+1 >= t.num_row() ) return 0 ;
  //if(t[j+k+1] <= x || t[j] > x) return 0 ;
  if( (t[j]>x) || (t[j+k+1]<x) || ( (j+k+1<t.num_row()-1) && t[j+k+1]<=x ) ) return 0 ;

  if(k==0) return 1 ;
  
  double v ;
  if(fabs(t[j+k]-t[j])< FUF) v = 0 ;
  else v = (x-t[j])/(t[k+j] - t[j]) * EvaluateBSpline(t, j, k-1, x) ;

  double w ;
  if(fabs(t[j+k+1]-t[j+1])< FUF) w = 0 ;
  else w = (t[j+k+1]-x)/(t[j+k+1]-t[j+1]) * EvaluateBSpline(t, j+1, k-1, x) ;
  
  return v + w ;
};

double
waxSplineParization::EvaluateBSplineDerivative(const HepVector& t,
					       const int j, const int k, 
					       const double x) 
{
  if(k==0) return 0 ;
  
  double vbar ;
  if(fabs(t[j+k]-t[j])< FUF) vbar = 0 ;
  else vbar = 1./(t[k+j] - t[j]) * EvaluateBSpline(t, j, k-1, x) ;

  double wbar ;
  if(fabs(t[j+k+1]-t[j+1])< FUF) wbar = 0 ;
  else wbar = 1./(t[j+k+1]-t[j+1]) * EvaluateBSpline(t, j+1, k-1, x) ; 
  return k*(vbar - wbar) ;
}


double
waxSplineParization::EvaluateBSplineSecondDerivative(const HepVector& t,
						     const int j, const int k, 
						     const double x) 
{
  if(k<=1) return 0 ;
  
  double vhat ;
  if(fabs(t[j+k]-t[j])< FUF) vhat = 0 ;
  else vhat = 1./(t[k+j] - t[j]) * 
         EvaluateBSplineDerivative(t, j, k-1, x) ;

  double what ;
  if(fabs(t[j+k+1]-t[j+1])< FUF) what = 0 ;
  else what = 1./(t[j+k+1]-t[j+1]) * 
         EvaluateBSplineDerivative(t, j+1, k-1, x) ; 
  return k*(vhat - what) ;
}

#undef FUF

