#include <fstream>
#include <iomanip>
#include "wax/RTRelation.hh"
#include "wax/Parization.hh"


//__________________________________________________________________
// waxRTRelation
//
// Interface class for the rt-relations
//
// Author: Wouter Hulsbergen
//__________________________________________________________________

waxRTRelation::DriftRadius_t
waxRTRelation::fgRMin[2] = {0.00125,0.00125 } ;//cm

waxRTRelation::DriftRadius_t
waxRTRelation::fgRMax[2] = {0.29,0.58} ;//cm

int waxRTRelation::fgNTableBins = 1000 ;

waxRTRelation::waxRTRelation() 
  : fTParization(0),fSParization(0),
    ftRTTable(0),ftMapTable(0) {}

#ifdef __OTR_USE_ROOT__
waxRTRelation::waxRTRelation(int key, EPitch pitch, int n_rt, int n_res,
			     TH1* histo,
			     DriftTime_t t_rmin, DriftTime_t t_rmax,
			     float resolution)
  : fLedaKey(key),fEPitch(pitch),fTParization(0),fSParization(0),
    ftRTTable(0),ftMapTable(0) 
{
  fTParization = waxParization::Create(GetRMin(),GetRMax(),n_rt,
				       histo,t_rmin,t_rmax) ;
  fSParization = waxParization::Create(GetRMin(),GetRMax(),n_res,
				       resolution*resolution) ;
}
#endif //__OTR_USE_ROOT__

waxRTRelation::waxRTRelation(const string& filename) 
  : fLedaKey(0),fTParization(0),fSParization(0),
    ftRTTable(0),ftMapTable(0)
{
  fTParization = waxParization::Create(GetRMin(),GetRMax(),0,0) ;
  fSParization = waxParization::Create(GetRMin(),GetRMax()) ;
  Read(filename) ;
}

waxRTRelation::waxRTRelation(LedaTableIterator<class OTR_r_t>& ledaIt)
  : fLedaKey(ledaIt->irt),fTParization(0),fSParization(0),
    ftRTTable(0),ftMapTable(0)
{
  fTParization = waxParization::Create(GetRMin(),GetRMax()) ;
  fSParization = waxParization::Create(GetRMin(),GetRMax()) ;
  ReadDB() ; 
}

waxRTRelation::~waxRTRelation() 
{
  delete fTParization ;
  delete fSParization ;
} 

void
waxRTRelation::ReadDB()
{
  LedaPointer<class OTR_r_t> ledartIt(OTR_r_t->get->key(fLedaKey)) ;
  fEPitch = ledartIt->pitch ;
  HepVector tpar(ledartIt->n_rt) ;
  for(int i=0; i<tpar.num_row(); i++) tpar[i] = ledartIt->par_rt[i] ;
  HepVector spar(ledartIt->n_res) ;
  for(int i=0; i<spar.num_row(); i++) spar[i] = ledartIt->par_res[i] ;
  fTParization = waxParization::Create(GetRMin(),GetRMax(),tpar) ;
  fSParization = waxParization::Create(GetRMin(),GetRMax(),spar) ;
  FillRTTable() ;
}

void
waxRTRelation::WriteDB() const
{
  LedaPointer<class OTR_r_t> ledartIt(OTR_r_t->add->key(fLedaKey)) ;
  //ledartIt->itr = fLedaKey ;
  ledartIt->pitch = fEPitch ;
  HepVector tpar = fTParization->GetPars() ;
  ledartIt->n_rt = tpar.num_row() ;
  for(int i=0; i<tpar.num_row(); i++) ledartIt->par_rt[i] = tpar[i] ;
  HepVector spar = fSParization->GetPars() ;
  ledartIt->n_res = spar.num_row() ;
  for(int i=0; i<spar.num_row(); i++) ledartIt->par_res[i] = spar[i] ;
}

void
waxRTRelation::Read(const string& filename) 
{
  ifstream ifs(filename.c_str()) ;
  if(ifs) {
    float rmin,rmax,t0;
    ifs >> rmin >> rmax >> t0;
    if(rmax>0.4) fEPitch=kPitchTen;
    else fEPitch=kPitchFive;
    fTParization->Read(GetRMin(),GetRMax(),ifs) ;
    fSParization->Read(GetRMin(),GetRMax(),ifs) ;
  } else {
    cout << "waxRTRelation::Read(...) : No such file " << filename << endl ;
  }
  FillRTTable() ;
}

void
waxRTRelation::Write(const string& filename) const 
{
  ofstream ofs(filename.c_str()) ;
  ofs << fEPitch << endl ;
  fTParization->Write(ofs) ;
  fSParization->Write(ofs) ;
}



waxRTRelation::RTResult
waxRTRelation:: CalcRTResultFromR(DriftRadius_t r) const
{
  RTResult rt;
  waxFunctionValue fv = fTParization->CalcFunction(r) ;
  rt.fR      = fv.fX ;
  rt.fT      = fv.fY ;
  rt.fdTdR   = fv.fdYdX ;
  rt.fSigmaR = sqrt(fSParization->CalcValue(r)) ;
  return rt;
}

waxRTRelation::RTResult
waxRTRelation:: CalcRTResultFromT(DriftTime_t t) const
{
  RTResult rt;
  float precision = 0.1 ; // channel
  waxFunctionValue fv = fTParization->CalcInverse(t,precision) ;
  rt.fR      = fv.fX ;
  rt.fT      = fv.fY ;
  rt.fdTdR   = fv.fdYdX ;
  rt.fSigmaR = sqrt(fSParization->CalcValue(rt.fR)) ;
  return rt;
}

void waxRTRelation::FillRTTable()
{  
  ftRTTable.resize(GetNTableBins()) ;
  ftMapTable.resize(GetNTableBins()) ;

  for(int rbin=0; rbin<GetNTableBins() ; rbin++) {
    DriftRadius_t r = GetRMin() + rbin*(GetRMax()-GetRMin())/GetNTableBins() ;
    ftRTTable[rbin] = CalcRTResultFromR(r) ;
  }
  
  // Fill inverse table
  ftTRMax = ftRTTable.back().fT ;
  ftTRMin = ftRTTable.front().fT ;
  float deltaT = (ftTRMax-ftTRMin)/fgNTableBins ;
  //bool positive = deltaT > 0 ;
  // too slow .... use sign deltaT !
  for(int tbin(0); tbin<fgNTableBins; tbin++) {
    float t = ftTRMin + tbin*deltaT ;
    int bestrbin=0;
    float bestdif=fabs(t - ftRTTable[0].fT) ;
    for(int rbin=1; rbin<fgNTableBins; rbin++) {
      float dif = fabs(t - ftRTTable[rbin].fT) ;
      if(dif < bestdif) {
	bestrbin = rbin;
	bestdif = dif;
      }
    }
    ftMapTable[tbin] = bestrbin ;
  }
}

